import pytest_mock
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome
from csle_attacker.emulation.util.shell_util import ShellUtil


class TestShellUtilSuite:
    """
    Test suite for shell_util.py
    """

    def test_execute_service_login_helper(self, mocker: pytest_mock.MockFixture,
                                          get_ex_em_env: EmulationEnvConfig) -> None:
        """
        Tests the execute_service_login_helper function

        :return: None
        """
        a = EmulationAttackerAction(
            id=EmulationAttackerActionId.CONTINUE, name="test", cmds=[], descr="", ips=["test"], index=0,
            action_outcome=EmulationAttackerActionOutcome.CONTINUE, vulnerability="", alt_cmds=[], backdoor=False,
            execution_time=0.0, ts=0.0, type=EmulationAttackerActionType.CONTINUE)
        s = EmulationEnvState(emulation_env_config=get_ex_em_env)
        mocker.patch('csle_common.util.connection_util.ConnectionUtil.login_service_helper',
                     return_value=(s, "conn"))
        s_prime = ShellUtil.execute_service_login_helper(s=s, a=a)
        for m in s_prime.attacker_obs_state.machines:
            if m.ips == a.ips:
                assert not m.untried_credentials



from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction


class AttackerStoppingMiddleware:
    """
    Class that implements optimal stopping actions for the attacker
    """

    @staticmethod
    def stop_intrusion(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a stopping action for the attacker

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        raise NotImplementedError("Not Implemented yet")

    @staticmethod
    def continue_intrusion(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a "continue" action for the attacker (does nothing)

        :param s: the current state
        :param a: the action to take
        :return: s_prime, reward, done
        """
        return s



from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
from csle_attacker.emulation.util.shell_util import ShellUtil


class PostExploitMiddleware:
    """
    Class that implements functionality for executing post-exploits actions on the emulation
    """

    @staticmethod
    def execute_service_login(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Executes a service login on the emulation using previously found credentials

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ShellUtil.execute_service_login_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_bash_find_flag(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Searches the file system for all servers where the agent is currently logged in to find flags

        :param s: the current state
        :param a: the action to take
        :return: s_prime, reward, done
        """
        new_machines_obs = []
        total_cost = 0.0
        for machine in s.attacker_obs_state.machines:
            if machine.logged_in:
                new_m_obs = EmulationAttackerMachineObservationState(ips=machine.ips)
                root_scan = False

                # Start with ssh connections
                new_m_obs, ssh_cost, ssh_root = ShellUtil._find_flag_using_ssh(
                    machine=machine, emulation_env_config=s.emulation_env_config, a=a, new_m_obs=new_m_obs)
                new_m_obs.filesystem_searched = True
                total_cost += ssh_cost
                if ssh_root:
                    root_scan = True

                if root_scan:
                    new_machines_obs.append(new_m_obs)
                    continue

                # If root scan not tried, try telnet connections
                new_m_obs, telnet_cost, telnet_root = ShellUtil._find_flag_using_telnet(
                    machine=machine, emulation_env_config=s.emulation_env_config, a=a, new_m_obs=new_m_obs)
                new_m_obs.filesystem_searched = True
                total_cost += telnet_cost
                if telnet_root:
                    root_scan = True

                if root_scan:
                    new_machines_obs.append(new_m_obs)
                    continue

                # If root scan not tried, try ftp connections
                new_m_obs, ftp_cost, ftp_root = ShellUtil._find_flag_using_ftp(
                    machine=machine, emulation_env_config=s.emulation_env_config, a=a, new_m_obs=new_m_obs)
                new_m_obs.filesystem_searched = True
                total_cost += ftp_cost
                if ftp_root:
                    root_scan = True
                new_machines_obs.append(new_m_obs)

        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, new_machines_obs, emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_install_tools(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Uses compromised machines with root access to install tools

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ShellUtil.install_tools_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_ssh_backdoor(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Uses compromised machines with root access to setup SSH backdoor

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ShellUtil.execute_ssh_backdoor_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime



from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_attacker.emulation.util.nmap_util import NmapUtil
from csle_attacker.emulation.util.nikto_util import NiktoUtil


class ReconMiddleware:
    """
    Class that implements functionality for executing reconnaissance actions on the emulation
    """

    @staticmethod
    def execute_tcp_syn_stealth_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a TCP SYN Stealth Scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_ping_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a Ping Scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_udp_port_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a UDP Port Scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_tcp_con_stealth_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a TCP CON Stealth scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_tcp_fin_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a TCP FIN scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_tcp_null_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a TCP Null scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_tcp_xmas_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a TCP Xmas scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_os_detection_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a OS detection scan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_vulscan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a vulscan action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_nmap_vulners(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a nmap_vulners scan

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_nikto_web_host_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a nikto web host scan

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NiktoUtil.nikto_scan_action_helper(s=s, a=a)

    @staticmethod
    def execute_masscan_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a masscan scan

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a, masscan=True)

    @staticmethod
    def execute_firewalk_scan(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a firewalk scan

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a, masscan=True)

    @staticmethod
    def execute_http_enum(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a http enum scan

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a, masscan=True)

    @staticmethod
    def execute_http_grep(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a http grep scan

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a, masscan=True)

    @staticmethod
    def execute_finger(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a finger scan

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        return NmapUtil.nmap_scan_action_helper(s=s, a=a, masscan=True)



from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_attacker.emulation.util.nmap_util import NmapUtil
from csle_attacker.emulation.util.exploit_util import ExploitUtil
from csle_common.logging.log import Logger


class ExploitMiddleware:
    """
    Class that implements functionality for executing exploits actions on the emulation
    """

    @staticmethod
    def execute_telnet_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a Telnet Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_ssh_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a SSH Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_ftp_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a FTP Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_cassandra_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a Cassandra Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_irc_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a IRC Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_mongo_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a Mongo Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_mysql_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a MySQL Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_smtp_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a SMTP Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_postgres_same_user_dictionary(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a Postgres Dictionary Password Attack action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = NmapUtil.nmap_scan_action_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_sambacry(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a Sambacry Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ExploitUtil.sambacry_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_shellshock(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a ShellShock Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ExploitUtil.shellshock_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_dvwa_sql_injection(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a DVWA SQL Injection Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ExploitUtil.dvwa_sql_injection_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_cve_2015_3306_exploit(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a CVE-2015-3306 Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ExploitUtil.cve_2015_3306_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_cve_2015_1427_exploit(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a CVE-2015-1427 Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ExploitUtil.cve_2015_1427_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_cve_2016_10033_exploit(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a CVE-2016-10033 Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        s_prime = ExploitUtil.cve_2016_10033_helper(s=s, a=a)
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_cve_2010_0426_exploit(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a CVE-2010-0426 Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        new_machines_obs = []
        total_cost = 0.0
        for machine in s.attacker_obs_state.machines:
            for ip in machine.ips:
                try:
                    if ip in a.ips and machine.logged_in and not machine.root:
                        new_m_obs = EmulationAttackerMachineObservationState(ips=machine.ips)
                        new_m_obs.cve_2010_0426_tried = True
                        # Try to escalate privileges using existing connections
                        new_m_obs, m_cost, exploit_successful, root_credential, service = \
                            ExploitUtil.cve_2010_0426_helper(s=s, a=a, machine=machine, result=new_m_obs)
                        new_machines_obs.append(new_m_obs.copy())

                        # Total cost update
                        total_cost += m_cost
                except Exception as e:
                    Logger.__call__().get_logger().warning(
                        f"There was an exception executing exploit CVE-2010-0426: {str(e)}, {repr(e)}")

        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, new_machines_obs, emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def execute_cve_2015_5602_exploit(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Performs a CVE-2015-5602 Exploit action

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        new_machines_obs = []
        total_cost = 0.0
        for machine in s.attacker_obs_state.machines:
            for ip in machine.ips:
                try:
                    if ip in a.ips and machine.logged_in and not machine.root:
                        new_m_obs = EmulationAttackerMachineObservationState(ips=machine.ips)
                        new_m_obs.cve_2015_5602_tried = True
                        # Try to escalate privileges using existing connections
                        new_m_obs, m_cost, exploit_successful, root_credential, service = \
                            ExploitUtil.cve_2015_5602_helper(
                                s=s, a=a, machine=machine, result=new_m_obs)
                        new_machines_obs.append(new_m_obs.copy())
                        # Total cost update
                        total_cost += m_cost
                except Exception as e:
                    Logger.__call__().get_logger().debug(f"{str(e)}, {repr(e)}")

        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, new_machines_obs, action=a, emulation_env_config=s.emulation_env_config)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime



from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_attacker.emulation.recon_middleware import ReconMiddleware
from csle_attacker.emulation.exploit_middleware import ExploitMiddleware
from csle_attacker.emulation.attacker_stopping_middleware import AttackerStoppingMiddleware
from csle_attacker.emulation.post_exploit_middleware import PostExploitMiddleware
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction


class EmulatedAttacker:
    """
    Represents an emulated attacker agent
    """

    @staticmethod
    def attacker_transition(s: EmulationEnvState, attacker_action: EmulationAttackerAction) -> EmulationEnvState:
        """
        Implements the transition operator T: (s,a) -> s'

        :param s: the current state
        :param attacker_action: the attacker action
        :param emulation_env_config: the emulation environment configuration
        :return: s'
        """
        if attacker_action.type == EmulationAttackerActionType.RECON:
            EnvDynamicsUtil.cache_attacker_action(a=attacker_action, s=s)
            return EmulatedAttacker.attacker_recon_action(s=s, a=attacker_action)
        elif attacker_action.type == EmulationAttackerActionType.EXPLOIT \
                or attacker_action.type == EmulationAttackerActionType.PRIVILEGE_ESCALATION:
            if attacker_action.index == -1:
                EnvDynamicsUtil.cache_attacker_action(a=attacker_action, s=s)
            return EmulatedAttacker.attacker_exploit_action(s=s, a=attacker_action)
        elif attacker_action.type == EmulationAttackerActionType.POST_EXPLOIT:
            return EmulatedAttacker.attacker_post_exploit_action(s=s, a=attacker_action)
        elif attacker_action.type == EmulationAttackerActionType.STOP \
                or attacker_action.type == EmulationAttackerActionType.CONTINUE:
            return EmulatedAttacker.attacker_stopping_action(s=s, a=attacker_action)
        else:
            raise ValueError("Action type not recognized")

    @staticmethod
    def attacker_recon_action(s: EmulationEnvState, a: EmulationAttackerAction) \
            -> EmulationEnvState:
        """
        Implements the transition of a reconnaissance action

        :param s: the current state
        :param a: the action
        :return: s'
        """
        if a.id == EmulationAttackerActionId.TCP_SYN_STEALTH_SCAN_HOST \
                or a.id == EmulationAttackerActionId.TCP_SYN_STEALTH_SCAN_ALL:
            return ReconMiddleware.execute_tcp_syn_stealth_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.PING_SCAN_HOST or a.id == EmulationAttackerActionId.PING_SCAN_ALL:
            return ReconMiddleware.execute_ping_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.UDP_PORT_SCAN_HOST \
                or a.id == EmulationAttackerActionId.UDP_PORT_SCAN_ALL:
            return ReconMiddleware.execute_udp_port_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.TCP_CON_NON_STEALTH_SCAN_HOST \
                or a.id == EmulationAttackerActionId.TCP_CON_NON_STEALTH_SCAN_ALL:
            return ReconMiddleware.execute_tcp_con_stealth_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.TCP_FIN_SCAN_HOST or a.id == EmulationAttackerActionId.TCP_FIN_SCAN_ALL:
            return ReconMiddleware.execute_tcp_fin_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.TCP_NULL_SCAN_HOST \
                or a.id == EmulationAttackerActionId.TCP_NULL_SCAN_ALL:
            return ReconMiddleware.execute_tcp_null_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.TCP_XMAS_TREE_SCAN_HOST \
                or a.id == EmulationAttackerActionId.TCP_XMAS_TREE_SCAN_ALL:
            return ReconMiddleware.execute_tcp_xmas_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.OS_DETECTION_SCAN_HOST \
                or a.id == EmulationAttackerActionId.OS_DETECTION_SCAN_ALL:
            return ReconMiddleware.execute_os_detection_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.VULSCAN_HOST \
                or a.id == EmulationAttackerActionId.VULSCAN_ALL:
            return ReconMiddleware.execute_vulscan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.NMAP_VULNERS_HOST  \
                or a.id == EmulationAttackerActionId.NMAP_VULNERS_ALL:
            return ReconMiddleware.execute_nmap_vulners(s=s, a=a)
        elif a.id == EmulationAttackerActionId.NIKTO_WEB_HOST_SCAN:
            return ReconMiddleware.execute_nikto_web_host_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.MASSCAN_HOST_SCAN or a.id == EmulationAttackerActionId.MASSCAN_ALL_SCAN:
            return ReconMiddleware.execute_masscan_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.FIREWALK_HOST \
                or a.id == EmulationAttackerActionId.FIREWALK_ALL:
            return ReconMiddleware.execute_firewalk_scan(s=s, a=a)
        elif a.id == EmulationAttackerActionId.HTTP_ENUM_HOST  \
                or a.id == EmulationAttackerActionId.HTTP_ENUM_ALL:
            return ReconMiddleware.execute_http_enum(s=s, a=a)
        elif a.id == EmulationAttackerActionId.HTTP_GREP_HOST \
                or a.id == EmulationAttackerActionId.HTTP_GREP_ALL:
            return ReconMiddleware.execute_http_grep(s=s, a=a)
        elif a.id == EmulationAttackerActionId.FINGER_HOST \
                or a.id == EmulationAttackerActionId.FINGER_ALL:
            return ReconMiddleware.execute_finger(s=s, a=a)
        else:
            raise ValueError("Recon action id:{},name:{} not recognized".format(a.id, a.name))

    @staticmethod
    def attacker_exploit_action(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Implements transition of an exploit action

        :param s: the current state
        :param a: the action
        :return: s'
        """
        if a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_telnet_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_ssh_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_ftp_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_cassandra_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_irc_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_mongo_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_mysql_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_smtp_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST \
                or a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL:
            return ExploitMiddleware.execute_postgres_same_user_dictionary(s=s, a=a)
        elif a.id == EmulationAttackerActionId.SAMBACRY_EXPLOIT:
            return ExploitMiddleware.execute_sambacry(s=s, a=a)
        elif a.id == EmulationAttackerActionId.SHELLSHOCK_EXPLOIT:
            return ExploitMiddleware.execute_shellshock(s=s, a=a)
        elif a.id == EmulationAttackerActionId.DVWA_SQL_INJECTION:
            return ExploitMiddleware.execute_dvwa_sql_injection(s=s, a=a)
        elif a.id == EmulationAttackerActionId.CVE_2015_3306_EXPLOIT:
            return ExploitMiddleware.execute_cve_2015_3306_exploit(s=s, a=a)
        elif a.id == EmulationAttackerActionId.CVE_2015_1427_EXPLOIT:
            return ExploitMiddleware.execute_cve_2015_1427_exploit(s=s, a=a)
        elif a.id == EmulationAttackerActionId.CVE_2016_10033_EXPLOIT:
            return ExploitMiddleware.execute_cve_2016_10033_exploit(s=s, a=a)
        elif a.id == EmulationAttackerActionId.CVE_2010_0426_PRIV_ESC:
            return ExploitMiddleware.execute_cve_2010_0426_exploit(s=s, a=a)
        elif a.id == EmulationAttackerActionId.CVE_2015_5602_PRIV_ESC:
            return ExploitMiddleware.execute_cve_2015_5602_exploit(s=s, a=a)
        else:
            raise ValueError("Exploit action id:{},name:{} not recognized".format(a.id, a.name))

    @staticmethod
    def attacker_post_exploit_action(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Implements the transition of a post-exploit action

        :param s: the current state
        :param a: the action
        :return: s'
        """
        if a.id == EmulationAttackerActionId.NETWORK_SERVICE_LOGIN:
            return PostExploitMiddleware.execute_service_login(s=s, a=a)
        if a.id == EmulationAttackerActionId.FIND_FLAG:
            return PostExploitMiddleware.execute_bash_find_flag(s=s, a=a)
        if a.id == EmulationAttackerActionId.INSTALL_TOOLS:
            return PostExploitMiddleware.execute_install_tools(s=s, a=a)
        if a.id == EmulationAttackerActionId.SSH_BACKDOOR:
            return PostExploitMiddleware.execute_ssh_backdoor(s=s, a=a)
        else:
            raise ValueError("Post-expoit action id:{},name:{} not recognized".format(a.id, a.name))

    @staticmethod
    def attacker_stopping_action(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Implements transition of a stopping action of the attacker

        :param s: the current state
        :param a: the action
        :return: s'
        """
        if a.id == EmulationAttackerActionId.STOP:
            return AttackerStoppingMiddleware.stop_intrusion(s=s, a=a)
        elif a.id == EmulationAttackerActionId.CONTINUE:
            return AttackerStoppingMiddleware.continue_intrusion(s=s, a=a)
        else:
            raise ValueError("Stopping action id:{},name:{} not recognized".format(a.id, a.name))



from typing import Tuple
import time
import paramiko
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state \
    import EmulationVulnerabilityObservationState
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential
from csle_common.util.emulation_util import EmulationUtil
from csle_common.util.connection_util import ConnectionUtil
from csle_common.logging.log import Logger


class ExploitUtil:
    """
    Class containing utility functions for the exploit-related functionality to the emulation
    """

    @staticmethod
    def sambacry_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Helper for executing the SambaCry exploit action

        :param s: the current state
        :param a: the SambaCry exploit action
        :return: s_prime
        """
        # Extract target machine
        target_machine = EmulationAttackerMachineObservationState(ips=a.ips)
        total_time = 0.0
        exploit_successful = False
        cost = 0.0
        for ip in target_machine.ips:
            try:
                jump_connection = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
                for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_SAMBACRY):
                    cost = 0
                    cmd = a.cmds[0]
                    cmd = cmd.format(ip)
                    outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=jump_connection.conn)
                    Logger.__call__().get_logger().warning(
                        f"Samba command: {cmd}, out: {outdata.decode()}, err: {errdata.decode()}")

                    # Parse result and check outcome
                    exploit_successful = (constants.SAMBA.ALREADY_EXISTS in outdata.decode() or
                                          (constants.SAMBA.ERROR not in outdata.decode() and
                                           constants.SAMBA.ERROR not in errdata.decode() and
                                           constants.SAMBA.AUTH_OK in outdata.decode()))

                    if exploit_successful:
                        break
                    else:
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_SAMBACRY_SLEEP_RETRY)
                if exploit_successful:
                    break
            except Exception as e:
                Logger.__call__().get_logger().warning(
                    f"There was an error executing the sambacry exploit: {str(e)} {repr(e)}")

        # Parse Result
        if exploit_successful:

            Logger.__call__().get_logger().info(f"The SambaCry exploit against machine: {target_machine.ips[0]} "
                                                f"was successful, cmd: {a.cmds[0].format(target_machine.ips[0])}")

            # Exploit successful
            credential = Credential(username=constants.SAMBA.BACKDOOR_USER, pw=constants.SAMBA.BACKDOOR_PW,
                                    port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                    service=constants.SSH.SERVICE_NAME)
            vuln = EmulationVulnerabilityObservationState(
                name=constants.EXPLOIT_VULNERABILITES.SAMBACRY_EXPLOIT,
                cvss=constants.EXPLOIT_VULNERABILITES.SAMBACRY_CVSS,
                credentials=[Credential(username=constants.SAMBA.BACKDOOR_USER, pw=constants.SAMBA.BACKDOOR_PW,
                                        port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                        service=constants.SSH.SERVICE_NAME)
                             ], port=constants.SAMBA.PORT, protocol=TransportProtocol.TCP,
                service=constants.SAMBA.SERVICE_NAME)
            target_machine.shell_access = True
            target_machine.untried_credentials = True
            target_machine.shell_access_credentials.append(credential)
            target_machine.backdoor_credentials.append(credential)
            target_machine.backdoor_tried = True
            target_machine.backdoor_installed = True
            target_machine.cve_vulns.append(vuln)

            # Measure  cost and alerts
            cost += float(total_time)

            EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                                   emulation_env_config=s.emulation_env_config)
        else:
            Logger.__call__().get_logger().info(f"The SambaCry exploit against machine: {target_machine.ips[0]} "
                                                f"was unsuccessful")

        target_machine.sambacry_tried = True
        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, [target_machine.copy()],
            emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        return s_prime

    @staticmethod
    def shellshock_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Helper for executing the ShellShock exploit action

        :param s: the current state
        :param a: the Shellshock exploit action
        :return: s_prime
        """
        target_machine = EmulationAttackerMachineObservationState(ips=a.ips)
        exploit_successful = False
        cost = 0.0
        total_time = 0.0
        for ip in target_machine.ips:
            try:
                jump_connection = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
                for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_SHELLSHOCK):
                    # Try execute exploit from jumphost
                    cost = 0
                    cmd = a.cmds[0]
                    cmd = cmd.format(constants.SHELLSHOCK.BACKDOOR_USER, constants.SHELLSHOCK.BACKDOOR_PW,
                                     constants.SHELLSHOCK.BACKDOOR_USER, ip)
                    outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=jump_connection.conn)
                    Logger.__call__().get_logger().info(f"Shellshock cmd: {cmd}, out: {outdata.decode()}, "
                                                        f"err:{errdata.decode()}")
                    cmd = a.cmds[1]
                    cmd = cmd.format(constants.SHELLSHOCK.BACKDOOR_USER, constants.SHELLSHOCK.BACKDOOR_PW, a.ips)
                    outdata1, errdata1, total_time1 = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=jump_connection.conn)
                    total_time = total_time + total_time1

                    # Parse Result
                    proxy_conn = EmulationConnectionObservationState(
                        conn=jump_connection.conn, credential=jump_connection.credential,
                        root=True, port=constants.SSH.DEFAULT_PORT, service=constants.SSH.SERVICE_NAME,
                        proxy=None, ip=s.emulation_env_config.containers_config.agent_ip)
                    exploit_successful = ExploitUtil.check_if_rce_exploit_succeeded(
                        user=constants.SHELLSHOCK.BACKDOOR_USER, pw=constants.SHELLSHOCK.BACKDOOR_PW,
                        source_ip=jump_connection.ip, port=constants.SSH.DEFAULT_PORT,
                        target_ip=ip, proxy_conn=proxy_conn)

                    if exploit_successful:
                        break
                    else:
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_SHELLSHOCK_SLEEP_RETRY)
                if exploit_successful:
                    break
            except Exception as e:
                Logger.__call__().get_logger().warning(f"There was an error executing the shellshock exploit: "
                                                       f"{str(e)}, {repr(e)}")
        if exploit_successful:

            Logger.__call__().get_logger().info(f"The Shellshock exploit against machine: {target_machine.ips[0]} "
                                                f"was successful")

            # Exploit successful
            credential = Credential(username=constants.SHELLSHOCK.BACKDOOR_USER,
                                    pw=constants.SHELLSHOCK.BACKDOOR_PW,
                                    port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                    service=constants.SSH.SERVICE_NAME)
            vuln = EmulationVulnerabilityObservationState(
                name=constants.EXPLOIT_VULNERABILITES.SHELLSHOCK_EXPLOIT,
                cvss=constants.EXPLOIT_VULNERABILITES.SHELLSHOCK_CVSS,
                credentials=[Credential(username=constants.SHELLSHOCK.BACKDOOR_USER,
                                        pw=constants.SHELLSHOCK.BACKDOOR_PW,
                                        port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                        service=constants.SSH.SERVICE_NAME)
                             ], port=constants.SHELLSHOCK.PORT, protocol=TransportProtocol.TCP,
                service=constants.SHELLSHOCK.SERVICE_NAME)
            target_machine.shell_access = True
            target_machine.untried_credentials = True
            target_machine.shell_access_credentials.append(credential)
            target_machine.backdoor_credentials.append(credential)
            target_machine.backdoor_tried = True
            target_machine.backdoor_installed = True
            target_machine.cve_vulns.append(vuln)
        else:
            Logger.__call__().get_logger().info(f"The Shellshock exploit against machine: {target_machine.ips[0]} "
                                                f"was unsuccessful")

        # Measure  cost and alerts
        cost += float(total_time)

        EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                               emulation_env_config=s.emulation_env_config)

        target_machine.shellshock_tried = True
        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, [target_machine.copy()],
            emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        return s_prime

    @staticmethod
    def check_if_rce_exploit_succeeded(user: str, pw: str, source_ip: str, port: int, target_ip: str,
                                       proxy_conn) -> bool:
        """
        Utility function for checking if a RCE exploit succeeded or not, i.e. if we got shell access

        :param user: the user that should have access
        :param pw: the password
        :param source_ip: the ip of the source node of the attack
        :param port: the port on the target
        :param target_ip: the target ip
        :param proxy_conn: optional proxy connection (jumphost)
        :return: True if the exploit succeeded, otherwise false
        """
        agent_addr = (source_ip, port)
        target_addr = (target_ip, port)
        agent_transport = proxy_conn.conn.get_transport()
        try:
            relay_channel = agent_transport.open_channel(constants.SSH.DIRECT_CHANNEL, target_addr, agent_addr,
                                                         timeout=3)
            target_conn = paramiko.SSHClient()
            target_conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            target_conn.connect(target_ip, username=user, password=pw, sock=relay_channel)
            transport = target_conn.get_transport()
            if transport is not None:
                transport.set_keepalive(5)
            return True
        except Exception as e:
            Logger.__call__().get_logger().debug(
                f"There was an error reading the result of an RCE exploit: {str(e)}, {repr(e)}")
            return False

    @staticmethod
    def dvwa_sql_injection_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Helper for executing the DVWA SQL Injection exploit action

        :param s: the current state
        :param a: the DVWA SQL Injection exploit action
        :return: s_prime
        """
        target_machine = EmulationAttackerMachineObservationState(ips=a.ips)
        cost = 0.0
        total_time = 0.0
        exploit_successful = False
        for ip in target_machine.ips:
            try:
                jump_connection = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
                for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_DVWA_SQL_INJECTION):
                    cmd = a.cmds[0]
                    cmd = cmd + " " + ip
                    outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=jump_connection.conn)
                    Logger.__call__().get_logger().info(f"Sql injeciton cmd: {cmd}, out: {outdata.decode()}, "
                                                        f"err:{errdata.decode()}")

                    # Parse Result
                    exploit_successful = False
                    dir = "/home/" + jump_connection.credential.username + "/"
                    try:
                        exploit_result = ExploitUtil.read_dvwa_sql_injection_result(conn=jump_connection.conn,
                                                                                    dir=dir)
                        pw = exploit_result[exploit_result.find("pablo:") + 6:
                                            exploit_result.find("pablo:") + 6 +
                                            exploit_result[exploit_result.find("pablo:") + 6:].find("<")]
                        if "0d107" in pw:
                            exploit_successful = True
                        else:
                            exploit_successful = False
                            pw = "-"
                    except Exception as e:
                        Logger.__call__().get_logger().warning(
                            f"There was an error executing the SQL injection exploit: {str(e)}, {repr(e)}")
                        exploit_successful = False
                        pw = "-"

                    if exploit_successful:
                        Logger.__call__().get_logger().info(f"SQL injection against: {target_machine.ips[0]} "
                                                            f"was successful")
                        ExploitUtil.remove_dvwa_sql_injection_result(conn=jump_connection.conn,
                                                                     dir=dir)
                        # Exploit successful
                        credential = Credential(username=constants.DVWA_SQL_INJECTION.EXPLOIT_USER,
                                                pw=pw,
                                                port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                                service=constants.SSH.SERVICE_NAME)
                        vuln = EmulationVulnerabilityObservationState(
                            name=constants.EXPLOIT_VULNERABILITES.DVWA_SQL_INJECTION,
                            cvss=9.8,
                            credentials=[Credential(username=constants.DVWA_SQL_INJECTION.EXPLOIT_USER,
                                                    pw=pw,
                                                    port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                                    service=constants.SSH.SERVICE_NAME)
                                         ], port=constants.DVWA_SQL_INJECTION.PORT, protocol=TransportProtocol.TCP,
                            service=constants.DVWA_SQL_INJECTION.SERVICE_NAME)
                        target_machine.shell_access = True
                        target_machine.untried_credentials = True
                        target_machine.shell_access_credentials.append(credential)
                        target_machine.backdoor_credentials.append(credential)
                        target_machine.backdoor_tried = True
                        target_machine.backdoor_installed = True
                        target_machine.cve_vulns.append(vuln)

                        # Measure  cost and alerts
                        cost += float(total_time)
                        break
                    else:
                        Logger.__call__().get_logger().info(f"SQL injection against: {target_machine.ips[0]} "
                                                            f"was not successful")
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_RETRY_DVWA_SQL_INJECTION)

                if exploit_successful:
                    break
            except Exception as e:
                Logger.__call__().get_logger().warning(
                    f"There was an error executing the SQL injection exploit: {e}, {repr(e)}")

        EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                               emulation_env_config=s.emulation_env_config)

        target_machine.dvwa_sql_injection_tried = True
        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, [target_machine.copy()],
            emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations

        return s_prime

    @staticmethod
    def read_dvwa_sql_injection_result(conn, dir: str) -> str:
        """
        Reads the result of a DVWA SQL Injection Attack
        :param conn: connection to use for reading
        :param dir: the dir to check
        :return: the read result
        """
        sftp_client = conn.open_sftp()
        remote_file = sftp_client.open(dir + constants.DVWA_SQL_INJECTION.EXPLOIT_OUTPUT_FILENAME)
        try:
            data = remote_file.read()
            result = data.decode()
        finally:
            remote_file.close()
        return str(result)

    @staticmethod
    def remove_dvwa_sql_injection_result(conn, dir: str) -> None:
        """
        Remove the result of a DVWA SQL Injection Attack
        :param conn: connection to use for the removal
        :param dir: the directory where the file is
        :return: None
        """
        sftp_client = conn.open_sftp()
        sftp_client.remove(dir + constants.DVWA_SQL_INJECTION.EXPLOIT_OUTPUT_FILENAME)

    @staticmethod
    def cve_2015_3306_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Helper for executing the CVE-2015-3306 exploit action

        :param s: the current state
        :param a: the CVE-2015-3306 exploit action
        :return: s_prime
        """
        # Extract target machine
        target_machine = EmulationAttackerMachineObservationState(ips=a.ips)
        exploit_successful = False
        cost = 0.0
        total_time = 0.0

        for ip in target_machine.ips:
            try:
                jump_connection = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
                for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CVE_2015_3306):
                    cost = 0
                    cmd = a.cmds[0]
                    cmd = cmd.format(a.ips)
                    outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=jump_connection.conn)
                    Logger.__call__().get_logger().info(f"CVE-2015-3306, cmd: {cmd}, out: {outdata.decode()}, "
                                                        f"err: {errdata.decode()}")
                    proxy_conn = EmulationConnectionObservationState(
                        conn=jump_connection.conn, credential=jump_connection.credential,
                        root=True, port=constants.SSH.DEFAULT_PORT, service=constants.SSH.SERVICE_NAME,
                        proxy=None, ip=s.emulation_env_config.containers_config.agent_ip)

                    exploit_successful = ExploitUtil.check_if_rce_exploit_succeeded(
                        user=constants.CVE_2015_3306.BACKDOOR_USER, pw=constants.CVE_2015_3306.BACKDOOR_PW,
                        source_ip=jump_connection.ip, port=constants.SSH.DEFAULT_PORT, target_ip=ip,
                        proxy_conn=proxy_conn)

                    if exploit_successful:
                        break
                    else:
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_CVE_2015_3306_SLEEP_RETRY)
                if exploit_successful:
                    break
            except Exception as e:
                Logger.__call__().get_logger().warning(
                    f"There was an error executing the CVE-2015-3306 exploit: {e}, {repr(e)}")

        if exploit_successful:
            Logger.__call__().get_logger().info(f"CVE-2015-3306 against target: {target_machine.ips[0]} "
                                                f"was successful")
            # Exploit successful
            credential = Credential(username=constants.CVE_2015_3306.BACKDOOR_USER,
                                    pw=constants.CVE_2015_3306.BACKDOOR_PW,
                                    port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                    service=constants.SSH.SERVICE_NAME)
            vuln = EmulationVulnerabilityObservationState(
                name=constants.EXPLOIT_VULNERABILITES.CVE_2015_3306,
                cvss=constants.EXPLOIT_VULNERABILITES.CVE_2015_3306_CVSS,
                credentials=[Credential(username=constants.CVE_2015_3306.BACKDOOR_USER,
                                        pw=constants.CVE_2015_3306.BACKDOOR_PW,
                                        port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                        service=constants.SSH.SERVICE_NAME)
                             ], port=constants.CVE_2015_3306.PORT, protocol=TransportProtocol.TCP,
                service=constants.CVE_2015_3306.SERVICE_NAME)
            target_machine.shell_access = True
            target_machine.untried_credentials = True
            target_machine.shell_access_credentials.append(credential)
            target_machine.backdoor_credentials.append(credential)
            target_machine.backdoor_tried = True
            target_machine.backdoor_installed = True
            target_machine.cve_vulns.append(vuln)
        else:
            Logger.__call__().get_logger().info(f"CVE-2015-3306 against target: {target_machine.ips[0]} "
                                                f"was not successful")

        # Measure  cost and alerts
        cost += float(total_time)

        EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                               emulation_env_config=s.emulation_env_config)
        target_machine.cve_2015_3306_tried = True

        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, [target_machine.copy()],
            emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        return s_prime

    @staticmethod
    def cve_2015_1427_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Helper for executing the CVE-2015-1427 exploit action

        :param s: the current state
        :param a: the CVE-2015-1427 exploit action
        :return: s_prime
        """
        # Extract target machine
        target_machine = EmulationAttackerMachineObservationState(ips=a.ips)
        cost = 0.0
        total_time = 0.0
        exploit_successful = False
        for ip in target_machine.ips:
            try:
                jump_connection = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
                for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CVE_2015_1427):
                    cost = 0
                    cmd = a.cmds[0]
                    cmd = cmd.format(a.ips)
                    outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=jump_connection.conn)
                    Logger.__call__().get_logger().info(f"CVE-2015-1427 cmd:{cmd}, out:{outdata.decode()}, "
                                                        f"err: {errdata.decode()}")

                    # Parse Result
                    proxy_conn = EmulationConnectionObservationState(
                        conn=jump_connection.conn, credential=jump_connection.credential,
                        root=True, port=constants.SSH.DEFAULT_PORT, service=constants.SSH.SERVICE_NAME,
                        proxy=None, ip=s.emulation_env_config.containers_config.agent_ip)

                    exploit_successful = ExploitUtil.check_if_rce_exploit_succeeded(
                        user=constants.CVE_2015_1427.BACKDOOR_USER, pw=constants.CVE_2015_1427.BACKDOOR_PW,
                        source_ip=jump_connection.ip, port=constants.SSH.DEFAULT_PORT, target_ip=ip,
                        proxy_conn=proxy_conn)

                    if exploit_successful:
                        break
                    else:
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_CVE_2015_1427_SLEEP_RETRY)
                if exploit_successful:
                    break
            except Exception as e:
                Logger.__call__().get_logger().warning(
                    f"There was an error executing the CVE-2015-1427 exploit: {e}, {repr(e)}")

        if exploit_successful:
            Logger.__call__().get_logger().info(f"CVE-2015-1427 against target: {target_machine.ips[0]} was successful")

            # Exploit successful
            credential = Credential(username=constants.CVE_2015_1427.BACKDOOR_USER,
                                    pw=constants.CVE_2015_1427.BACKDOOR_PW,
                                    port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                    service=constants.SSH.SERVICE_NAME)
            vuln = EmulationVulnerabilityObservationState(
                name=constants.EXPLOIT_VULNERABILITES.CVE_2015_1427,
                cvss=constants.EXPLOIT_VULNERABILITES.CVE_2015_1427_CVSS,
                credentials=[Credential(username=constants.CVE_2015_1427.BACKDOOR_USER,
                                        pw=constants.CVE_2015_1427.BACKDOOR_PW,
                                        port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                        service=constants.SSH.SERVICE_NAME)
                             ], port=constants.CVE_2015_1427.PORT, protocol=TransportProtocol.TCP,
                service=constants.CVE_2015_1427.SERVICE_NAME)
            target_machine.shell_access = True
            target_machine.untried_credentials = True
            target_machine.shell_access_credentials.append(credential)
            target_machine.backdoor_credentials.append(credential)
            target_machine.backdoor_tried = True
            target_machine.backdoor_installed = True
            target_machine.cve_vulns.append(vuln)
        else:
            Logger.__call__().get_logger().info(f"CVE-2015-1427 against target: {target_machine.ips[0]} "
                                                f"was not successful")

        # Measure  cost and alerts
        cost += float(total_time)
        EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                               emulation_env_config=s.emulation_env_config)
        target_machine.cve_2015_1427_tried = True

        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, [target_machine.copy()],
            emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        return s_prime

    @staticmethod
    def cve_2016_10033_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Helper for executing the CVE-2016-10033 exploit action

        :param s: the current state
        :param a: the CVE-2016-10033 exploit action
        :return: s_prime
        """
        # Extract target machine
        target_machine = EmulationAttackerMachineObservationState(ips=a.ips)
        cost = 0.0
        total_time = 0.0
        exploit_successful = False
        for ip in target_machine.ips:
            try:
                jump_connection = ConnectionUtil.find_jump_host_connection(ip=ip, s=s,
                                                                           emulation_env_config=s.emulation_env_config)
                for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CVE_2016_10033):
                    cost = 0
                    cmd = a.cmds[0]
                    cmd = cmd.format(a.ips)
                    outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=jump_connection.conn)
                    Logger.__call__().get_logger().info(f"CVE-2016-10033 cmd:{cmd}, out: {outdata.decode()}, "
                                                        f"err: {errdata.decode()}")

                    # Parse Result
                    proxy_conn = EmulationConnectionObservationState(
                        conn=jump_connection.conn, credential=jump_connection.credential,
                        root=True, port=constants.SSH.DEFAULT_PORT, service=constants.SSH.SERVICE_NAME,
                        proxy=None, ip=s.emulation_env_config.containers_config.agent_ip)

                    exploit_successful = ExploitUtil.check_if_rce_exploit_succeeded(
                        user=constants.CVE_2016_10033.BACKDOOR_USER, pw=constants.CVE_2016_10033.BACKDOOR_PW,
                        source_ip=jump_connection.ip, port=constants.SSH.DEFAULT_PORT, target_ip=ip,
                        proxy_conn=proxy_conn)

                    if exploit_successful:
                        break
                    else:
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_CVE_2016_10033_SLEEP_RETRY)
                if exploit_successful:
                    break
            except Exception as e:
                Logger.__call__().get_logger().warning(
                    f"There was an exception executing exploit CVE-2016-10033: {e}, {repr(e)}")

        if exploit_successful:
            Logger.__call__().get_logger().info(f"CVE-2016-10033 against {target_machine.ips[0]} was succcessful")

            # Exploit successful
            credential = Credential(username=constants.CVE_2016_10033.BACKDOOR_USER,
                                    pw=constants.CVE_2016_10033.BACKDOOR_PW,
                                    port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                    service=constants.SSH.SERVICE_NAME)
            vuln = EmulationVulnerabilityObservationState(
                name=constants.EXPLOIT_VULNERABILITES.CVE_2016_10033,
                cvss=constants.EXPLOIT_VULNERABILITES.CVE_2016_10033_CVSS,
                credentials=[Credential(username=constants.CVE_2016_10033.BACKDOOR_USER,
                                        pw=constants.CVE_2016_10033.BACKDOOR_PW,
                                        port=constants.SSH.DEFAULT_PORT, protocol=TransportProtocol.TCP,
                                        service=constants.SSH.SERVICE_NAME)
                             ], port=constants.CVE_2016_10033.PORT, protocol=TransportProtocol.TCP,
                service=constants.CVE_2016_10033.SERVICE_NAME)
            target_machine.shell_access = True
            target_machine.untried_credentials = True
            target_machine.shell_access_credentials.append(credential)
            target_machine.backdoor_credentials.append(credential)
            target_machine.backdoor_tried = True
            target_machine.backdoor_installed = True
            target_machine.cve_vulns.append(vuln)
        else:
            Logger.__call__().get_logger().info(f"CVE-2016-10033 against {target_machine.ips[0]} was not succcessful")

        # Measure  cost and alerts
        cost += float(total_time)
        EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                               emulation_env_config=s.emulation_env_config)

        target_machine.cve_2016_10033_tried = True

        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, [target_machine.copy()],
            emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations

        return s_prime

    @staticmethod
    def cve_2010_0426_helper(s: EmulationEnvState, a: EmulationAttackerAction,
                             machine: EmulationAttackerMachineObservationState,
                             result: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, float, bool, Credential, str]:
        """
        Helper for executing the CVE-2010-0426 privelege_escalation action

        :param s: the current state
        :param a: the CVE-2010-0425 privilege escalation action
        :return: new_machine_obs, cost, successful, root_credential, service_name
        """
        cost = 0.0
        e_succ = False
        root_credential = None
        service = ""
        sleep_time = 0.5

        # Start with SSH connections
        for ssh_conn in machine.ssh_connections:
            for ip in a.ips:
                try:
                    proxy_conn = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
                    for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CVE_2010_0426):

                        # Setup interactive shell
                        channel = ssh_conn.conn.invoke_shell()

                        # clear output
                        if channel.recv_ready():
                            channel.recv(constants.COMMON.DEFAULT_RECV_SIZE)

                        # Exploit
                        cmd = a.cmds[0].format(constants.CVE_2010_0426.EXPLOIT_FILE)
                        e_succ = False
                        try:
                            EmulationUtil.execute_cmd_interactive_channel(cmd=cmd, channel=channel)
                            time.sleep(sleep_time)
                            EmulationUtil.read_result_interactive_channel(
                                emulation_env_config=s.emulation_env_config, channel=channel)
                            cmd = a.cmds[1]
                            EmulationUtil.execute_cmd_interactive_channel(cmd=cmd, channel=channel)
                            time.sleep(sleep_time)
                            EmulationUtil.read_result_interactive_channel(
                                emulation_env_config=s.emulation_env_config, channel=channel)
                            e_succ = ExploitUtil.check_if_rce_exploit_succeeded(
                                user=constants.CVE_2010_0426.BACKDOOR_USER, pw=constants.CVE_2010_0426.BACKDOOR_PW,
                                source_ip=s.emulation_env_config.containers_config.agent_ip,
                                port=constants.SSH.DEFAULT_PORT,
                                target_ip=ip, proxy_conn=proxy_conn)
                        except Exception as e:
                            Logger.__call__().get_logger().warning("CVE-2010-0426 error:{}".format(str(e)))
                            e_succ = False

                        if e_succ:
                            Logger.__call__().get_logger().info("CVE-2010-0426 was succcessful")
                            root_credential = Credential(
                                username=constants.CVE_2010_0426.BACKDOOR_USER,
                                pw=constants.CVE_2010_0426.BACKDOOR_PW, port=constants.SSH.DEFAULT_PORT,
                                protocol=TransportProtocol.TCP, service=constants.SSH.SERVICE_NAME)
                            root_connection = EmulationUtil.setup_custom_connection(
                                user=constants.CVE_2010_0426.BACKDOOR_USER, pw=constants.CVE_2010_0426.BACKDOOR_PW,
                                source_ip=s.emulation_env_config.containers_config.agent_ip,
                                port=constants.SSH.DEFAULT_PORT,
                                target_ip=ip, proxy_conn=proxy_conn, root=True)
                            vuln = EmulationVulnerabilityObservationState(
                                name=constants.EXPLOIT_VULNERABILITES.CVE_2010_0426,
                                cvss=constants.EXPLOIT_VULNERABILITES.CVE_2010_0426_CVSS,
                                credentials=[root_credential], port=constants.SSH.DEFAULT_PORT,
                                protocol=TransportProtocol.TCP,
                                service=root_credential.service)
                            result.cve_vulns.append(vuln)
                            result.cve_2010_0426_tried = True
                            result.shell_access_credentials.append(root_credential)
                            result.backdoor_credentials.append(root_credential)
                            result.backdoor_tried = True
                            result.backdoor_installed = True
                            result.shell_access = True
                            result.logged_in = True
                            result.root = True
                            result.ssh_connections.append(root_connection)
                            service = constants.SSH.SERVICE_NAME
                            break
                        else:
                            Logger.__call__().get_logger().info("CVE-2010-0426 against was not succcessful")
                            time.sleep(constants.ENV_CONSTANTS.ATTACKER_CVE_2010_0426_SLEEP_RETRY)
                    if e_succ:
                        break
                except Exception as e:
                    Logger.__call__().get_logger().warning(
                        f"Exception occurred during execution of exploit CVE-2010-0426, {str(e)}, {repr(e)}")

        # Measure  cost and alerts
        cost += float(sleep_time * 2)
        EmulationUtil.log_measured_action_time(total_time=cost, action=a, emulation_env_config=s.emulation_env_config)
        result.cve_2010_0426_tried = True
        return result, cost, e_succ, root_credential, service

    @staticmethod
    def cve_2015_5602_helper(s: EmulationEnvState, a: EmulationAttackerAction,
                             machine: EmulationAttackerMachineObservationState,
                             result: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, float, bool, Credential, str]:
        """
        Helper for executing the CVE-2015-5602 privilege escalation action

        :param s: the current state
        :param a: the CVE-2015-5602 privilege escalation action
        :return: s_prime, cost, e_succ (flag), credentials, servicename
        """
        cost = 0.0
        e_succ = False
        root_credential = None
        service = ""
        sleep_time = 0.5

        # Start with SSH connections
        for ssh_conn in machine.ssh_connections:
            for ip in a.ips:
                proxy_conn = ConnectionUtil.find_jump_host_connection(ip=ip, s=s)
                for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CVE_2015_5602):

                    # Setup interactive shell
                    channel = ssh_conn.conn.invoke_shell()

                    # clear output
                    if channel.recv_ready():
                        channel.recv(constants.COMMON.DEFAULT_RECV_SIZE)

                    # Exploit
                    cmd = a.cmds[0]
                    try:
                        EmulationUtil.execute_cmd_interactive_channel(cmd=cmd, channel=channel)
                        time.sleep(sleep_time)
                        cost += float(sleep_time)
                        EmulationUtil.read_result_interactive_channel(
                            emulation_env_config=s.emulation_env_config, channel=channel)
                        cmd = a.cmds[1]
                        EmulationUtil.execute_cmd_interactive_channel(cmd=cmd, channel=channel)
                        time.sleep(sleep_time)
                        cost += float(sleep_time)
                        EmulationUtil.read_result_interactive_channel(
                            emulation_env_config=s.emulation_env_config, channel=channel)
                        cmd = a.cmds[2]
                        EmulationUtil.execute_cmd_interactive_channel(cmd=cmd, channel=channel)
                        time.sleep(sleep_time)
                        cost += float(sleep_time)
                        EmulationUtil.read_result_interactive_channel(
                            emulation_env_config=s.emulation_env_config, channel=channel)
                        cmd = a.cmds[3]
                        EmulationUtil.execute_cmd_interactive_channel(cmd=cmd, channel=channel)
                        time.sleep(sleep_time)
                        cost += float(sleep_time)
                        EmulationUtil.read_result_interactive_channel(
                            emulation_env_config=s.emulation_env_config, channel=channel)
                        e_succ = ExploitUtil.check_if_rce_exploit_succeeded(
                            user=constants.CVE_2015_5602.BACKDOOR_USER, pw=constants.CVE_2015_5602.BACKDOOR_PW,
                            source_ip=s.emulation_env_config.containers_config.agent_ip,
                            port=constants.SSH.DEFAULT_PORT,
                            target_ip=ip, proxy_conn=proxy_conn)
                    except Exception as e:
                        Logger.__call__().get_logger().warning(f"CVE-2015-5602 error:{str(e)}, {repr(e)}")
                        e_succ = False

                    if e_succ:
                        Logger.__call__().get_logger().info("CVE-2015-5602 was succcessful")
                        root_credential = Credential(username=constants.CVE_2015_5602.BACKDOOR_USER,
                                                     pw=constants.CVE_2015_5602.BACKDOOR_PW,
                                                     port=constants.SSH.DEFAULT_PORT,
                                                     protocol=TransportProtocol.TCP,
                                                     service=constants.SSH.SERVICE_NAME)
                        root_connection = EmulationUtil.setup_custom_connection(
                            user=constants.CVE_2015_5602.BACKDOOR_USER, pw=constants.CVE_2015_5602.BACKDOOR_PW,
                            source_ip=s.emulation_env_config.containers_config.agent_ip,
                            port=constants.SSH.DEFAULT_PORT,
                            target_ip=ip, proxy_conn=proxy_conn, root=True)
                        vuln = EmulationVulnerabilityObservationState(
                            name=constants.EXPLOIT_VULNERABILITES.CVE_2015_5602,
                            cvss=constants.EXPLOIT_VULNERABILITES.CVE_2015_5602_CVSS,
                            credentials=[root_credential], port=constants.SSH.DEFAULT_PORT,
                            protocol=TransportProtocol.TCP, service=root_credential.service)
                        result.cve_vulns.append(vuln)
                        result.cve_2015_5602_tried = True
                        result.shell_access_credentials.append(root_credential)
                        result.backdoor_credentials.append(root_credential)
                        result.backdoor_tried = True
                        result.backdoor_installed = True
                        result.shell_access = True
                        result.logged_in = True
                        result.root = True
                        result.ssh_connections.append(root_connection)
                        service = constants.SSH.SERVICE_NAME
                    else:
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_CVE_2015_5602_SLEEP_RETRY)
                        cost += float(constants.ENV_CONSTANTS.ATTACKER_CVE_2015_5602_SLEEP_RETRY)

            # Measure  cost and alerts
            cost += float(sleep_time * 2)
            EmulationUtil.log_measured_action_time(total_time=cost, action=a,
                                                   emulation_env_config=s.emulation_env_config)
        result.cve_2015_5602_tried = True
        return result, cost, e_succ, root_credential, service



from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_attacker.emulation.emulated_attacker import EmulatedAttacker


class Attacker:
    """
    Represents an automated attacker agent
    """

    @staticmethod
    def attacker_transition(s: EmulationEnvState, attacker_action: EmulationAttackerAction) -> EmulationEnvState:
        """
        Implements an attacker transition of the MDP/Markov Game:
        (s, a) --> s'

        :param s: the current environment state in the emulation
        :param attacker_action: the attacker action
        :param simulation: boolean flag if it is a simulated transition or an actual transition in the emulation
        :return: s' (EmulationEnvState)
        """
        return EmulatedAttacker.attacker_transition(s=s, attacker_action=attacker_action)



import xml.etree.ElementTree as ET
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action_result.nikto_scan_result \
    import NiktoScanResult
from csle_common.dao.emulation_action_result.nikto_vuln import NiktoVuln
from csle_common.util.emulation_util import EmulationUtil
from csle_attacker.emulation.util.nmap_util import NmapUtil
from csle_common.logging.log import Logger


class NiktoUtil:
    """
    Class containing utility functions for the nikto-related functionality to the emulation
    """

    @staticmethod
    def parse_nikto_scan(file_name: str, emulation_env_config: EmulationEnvConfig) -> ET.Element:
        """
        Parses an XML file containing the result of an nikt scan

        :param file_name: name of the file to parse
        :param emulation_env_config: environment config
        :return: the parsed xml file
        """
        sftp_client = emulation_env_config.get_hacker_connection().open_sftp()
        remote_file = sftp_client.open(constants.NMAP.RESULTS_DIR + file_name)
        try:
            xml_tree = ET.parse(remote_file)
        finally:
            remote_file.close()
        xml_data = xml_tree.getroot()
        return xml_data

    @staticmethod
    def nikto_scan_action_helper(s: EmulationEnvState, a: EmulationAttackerAction) \
            -> EmulationEnvState:
        """
        Helper function for executing a NIKTO web scan action on the emulation. Implements caching.

        :param s: the current env state
        :param a: the Nikto action to execute
        :param emulation_env_config: the emulation env config
        :return: s', reward, done
        """
        cmds, file_names = a.nikto_cmds()
        outdata, errdata, total_time = EmulationUtil.execute_ssh_cmds(
            cmds=cmds, conn=s.emulation_env_config.get_hacker_connection())
        EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                               emulation_env_config=s.emulation_env_config)

        # Read result
        scan_result = NiktoScanResult(ip=a.ips[0], vulnerabilities=[], port=80, sitename=a.ips[0])
        for file_name in file_names:
            for i in range(constants.ENV_CONSTANTS.NUM_RETRIES):
                try:
                    xml_data = NiktoUtil.parse_nikto_scan(
                        file_name=file_name, emulation_env_config=s.emulation_env_config)
                    scan_result = NiktoUtil.parse_nikto_scan_xml(xml_data)
                    s = NiktoUtil.merge_nikto_scan_result_with_state(
                        scan_result=scan_result, s=s, a=a)
                    break
                except Exception as e:
                    Logger.__call__().get_logger().warning(
                        f"There was an error parsing the Nikto scan output: {e}, {repr(e)}")
        s_prime = s
        return s_prime

    @staticmethod
    def merge_nikto_scan_result_with_state(scan_result: NiktoScanResult, s: EmulationEnvState,
                                           a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Merges a Nikto scan result with an existing observation state

        :param scan_result: the scan result
        :param s: the current state
        :param a: the action just executed
        :return: s', reward, done
        """
        m_obs = None

        for m in s.attacker_obs_state.machines:
            if m.ips == scan_result.ip:
                m_obs = EmulationAttackerMachineObservationState(ips=m.ips)

        if m_obs is None:
            raise ValueError(f"Unknown IP: {scan_result.ip}")
        else:
            for vuln in scan_result.vulnerabilities:
                vuln_obs = vuln.to_obs()
                m_obs.osvdb_vulns.append(vuln_obs)

            attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
                s.attacker_obs_state.machines, [m_obs], emulation_env_config=s.emulation_env_config, action=a)
            s_prime = s
            s_prime.attacker_obs_state.machines = attacker_machine_observations

            return s_prime

    @staticmethod
    def parse_nikto_scan_xml(xml_data) -> NiktoScanResult:
        """
        Parses an XML Tree with Nikto Scan Result into a Nikto Scan DTO

        :param xml_data: the xml tree of Nikto Scan Result to parse
        :return: parsed nikto scan result
        """
        result = None
        for child in xml_data:
            if child.tag == constants.NIKTO_XML.SCANDETAILS:
                result = NiktoUtil._parse_nikto_scandetails(child)
            elif child.tag == constants.NIKTO_XML.ITEM:
                result = NiktoUtil._parse_nikto_scandetails(xml_data)
            elif child.tag == constants.NIKTO_XML.NIKTOSCAN:
                NiktoUtil.parse_nikto_scan_xml(xml_data)
        return result

    @staticmethod
    def _parse_nikto_scandetails(xml_data) -> NiktoScanResult:
        """
        Parses a host-element in the XML tree

        :param xml_data: the host element
        :return: parsed nikto scan result
        """
        target_ip = ""
        targetport = ""
        vulnerabilities = []

        if constants.NIKTO_XML.TARGETPORT in xml_data.keys():
            targetport = xml_data.attrib[constants.NIKTO_XML.TARGETPORT]
        if constants.NIKTO_XML.TARGETIP in xml_data.keys():
            target_ip = xml_data.attrib[constants.NIKTO_XML.TARGETIP]

        for child in list(xml_data.iter()):
            if child.tag == constants.NIKTO_XML.ITEM:
                vuln = NiktoUtil._parse_nikto_item(NmapUtil._parse_nmap_status_xml(child))
                vulnerabilities.append(vuln)
        nikto_scan_result = NiktoScanResult(vulnerabilities=vulnerabilities, port=targetport,
                                            ip=target_ip, sitename=target_ip)
        return nikto_scan_result

    @staticmethod
    def _parse_nikto_item(xml_data) -> NiktoVuln:
        """
        Parses a item in the XML tree of a Nikto scan

        :param xml_data: the item element
        :return: parsed nikto vuln
        """
        id = ""
        osvdb_id = None
        method = ""
        iplink = ""
        namelink = ""
        uri = ""
        description = ""

        if constants.NIKTO_XML.METHOD in xml_data.keys():
            method = xml_data.attrib[constants.NIKTO_XML.METHOD]
        if constants.NIKTO_XML.OSVDB_ID in xml_data.keys():
            osvdb_id = int(xml_data.attrib[constants.NIKTO_XML.OSVDB_ID])
        if constants.NIKTO_XML.ITEM_ID in xml_data.keys():
            id = str(int(xml_data.attrib[constants.NIKTO_XML.ITEM_ID]))

        for child in list(xml_data.iter()):
            if child.tag == constants.NIKTO_XML.DESCR:
                description = child.text
            elif child.tag == constants.NIKTO_XML.URI:
                uri = child.text
            elif child.tag == constants.NIKTO_XML.NAMELINK:
                namelink = child.text
            elif child.tag == constants.NIKTO_XML.IPLINK:
                iplink = child.text

        nikto_vuln = NiktoVuln(id=id, osvdb_id=osvdb_id, method=method, iplink=iplink, namelink=namelink,
                               uri=uri, description=description)

        return nikto_vuln



from typing import Tuple, List, Union, Any
import time
import threading
import xml.etree.ElementTree as ET
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action \
    import EmulationAttackerAction
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action_result.nmap_scan_result import NmapScanResult
from csle_common.dao.emulation_action_result.nmap_host_result import NmapHostResult
from csle_common.dao.emulation_action_result.nmap_port_status import NmapPortStatus
from csle_common.dao.emulation_action_result.nmap_host_status import NmapHostStatus
from csle_common.dao.emulation_action_result.nmap_port import NmapPort
from csle_common.dao.emulation_action_result.nmap_addr_type import NmapAddrType
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_action_result.nmap_os import NmapOs
from csle_common.dao.emulation_action_result.nmap_vuln import NmapVuln
from csle_common.dao.emulation_action_result.nmap_brute_credentials import NmapBruteCredentials
from csle_common.dao.emulation_action_result.nmap_hop import NmapHop
from csle_common.dao.emulation_action_result.nmap_trace import NmapTrace
from csle_common.dao.emulation_action_result.nmap_http_enum import NmapHttpEnum
from csle_common.dao.emulation_action_result.nmap_http_grep import NmapHttpGrep
from csle_common.dao.emulation_action_result.nmap_vulscan import NmapVulscan
from csle_common.util.emulation_util import EmulationUtil
from csle_common.util.connection_util import ConnectionUtil
from csle_common.logging.log import Logger


class NmapUtil:
    """
    Class containing utility functions for the nmap-related functionality to the emulation
    """

    @staticmethod
    def parse_nmap_scan(file_name: str, emulation_env_config: EmulationEnvConfig, conn=None, dir: str = "") \
            -> ET.Element:
        """
        Parses an XML file containing the result of an nmap scan

        :param file_name: name of the file to parse
        :param conn: the SSH connection to use for parsing
        :param dir: the directory to parse the XML file
        :param emulation_env_config: environment config
        :return: the parsed xml file
        """
        if conn is None:
            conn = emulation_env_config.get_hacker_connection()
        if dir is None or dir == "":
            dir = constants.NMAP.RESULTS_DIR

        sftp_client = conn.open_sftp()
        file_name = str(dir + file_name).strip()
        remote_file = sftp_client.open(file_name)
        try:
            xml_tree = ET.parse(remote_file)
        finally:
            remote_file.close()
        xml_data = xml_tree.getroot()
        return xml_data

    @staticmethod
    def parse_nmap_scan_xml(xml_data, ips: List[str], action: EmulationAttackerAction) -> NmapScanResult:
        """
        Parses an XML Tree into a DTO

        :param xml_data: the xml tree to parse
        :param ips: ips of the source of the scan
        :param action: the action of the scan
        :return: parsed nmap scan result
        """
        hosts: List[NmapHostResult] = []
        for child in xml_data:
            if child.tag == constants.NMAP_XML.HOST:
                host = NmapUtil._parse_nmap_host_xml(child, action=action)
                hosts = NmapUtil._merge_nmap_hosts(host, hosts)
        result = NmapScanResult(hosts=hosts, ips=ips)
        return result

    @staticmethod
    def _parse_nmap_host_xml(xml_data, action: EmulationAttackerAction) -> NmapHostResult:
        """
        Parses a host-element in the XML tree

        :param xml_data: the host element
        :param action: action of the scan
        :return: parsed nmap host result
        """
        ip_addr = None
        mac_addr = None
        hostnames = []
        ports: List[NmapPort] = []
        vulnerabilities: List[NmapVuln] = []
        credentials: List[NmapBruteCredentials] = []
        os = None
        os_matches = []
        status = NmapHostStatus.UP
        trace = None
        for child in list(xml_data.iter()):
            if child.tag == constants.NMAP_XML.ADDRESS:
                addr, type = NmapUtil._parse_nmap_address_xml(child)
                if type == NmapAddrType.MAC:
                    mac_addr = addr
                else:
                    ip_addr = addr
            elif child.tag == constants.NMAP_XML.HOSTNAMES:
                hostnames = NmapUtil._parse_nmap_hostnames_xml(child)
            elif child.tag == constants.NMAP_XML.PORTS:
                ports, vulnerabilities, credentials = NmapUtil._parse_nmap_ports_xml(child, action=action)
            elif child.tag == constants.NMAP_XML.OS:
                os_matches = NmapUtil._parse_nmap_os_xml(child)
                os = NmapOs.get_best_match(os_matches)
            elif child.tag == constants.NMAP_XML.TRACE:
                trace = NmapUtil._parse_nmap_trace_xml(child)
        nmap_host_result = NmapHostResult(status=status, ips=[ip_addr], mac_addr=mac_addr,
                                          hostnames=hostnames, ports=ports, os=os, os_matches=os_matches,
                                          vulnerabilities=vulnerabilities, credentials=credentials, trace=trace)
        return nmap_host_result

    @staticmethod
    def _parse_nmap_status_xml(xml_data) -> NmapPortStatus:
        """
        Parses a status element in the xml tree

        :param xml_data: the status XML element
        :return: parsed Nmap Port-Status DTO
        """
        status = NmapPortStatus.DOWN
        status_val = xml_data.attrib[constants.NMAP_XML.STATE]
        if status_val == constants.NMAP_XML.STATUS_UP:
            status = NmapPortStatus.UP
        return status

    @staticmethod
    def _parse_nmap_address_xml(xml_data) -> Tuple[str, NmapAddrType]:
        """
        Parses a address element in the xml tree

        :param xml_data: the address XML element
        :return: (address, addresstype)
        """
        type = NmapAddrType.IP
        addr = xml_data.attrib[constants.NMAP_XML.ADDR]
        addrtype_val = xml_data.attrib[constants.NMAP_XML.ADDR_TYPE]
        if constants.NMAP_XML.IP in addrtype_val:
            type = NmapAddrType.IP
        elif constants.NMAP_XML.MAC in addrtype_val:
            type = NmapAddrType.MAC
        return addr, type

    @staticmethod
    def _parse_nmap_hostnames_xml(xml_data) -> List[str]:
        """
        Parses a hostnames element in the XML tree

        :param xml_data: the hostnames XML element
        :return: a list of hostnames
        """
        hostnames = []
        for child in list(xml_data.iter()):
            if child.tag == constants.NMAP_XML.HOSTNAME:
                hostnames.append(child.attrib[constants.NMAP_XML.NAME])
        return hostnames

    @staticmethod
    def _parse_nmap_ports_xml(xml_data, action: EmulationAttackerAction) \
            -> Tuple[List[NmapPort], List[NmapVuln], List[NmapBruteCredentials]]:
        """
        Parses a ports XML element in the XML tree

        :param xml_data: the ports XML element
        :param action: action of the scan
        :return: (List NmapPort, List NmapVuln, ListNmapBruteCredentials)
        """
        ports = []
        vulnerabilities = []
        credentials = []
        http_enum = None
        http_grep = None
        vulscan = None
        for child in list(xml_data.iter()):
            if child.tag == constants.NMAP_XML.PORT:
                port_status = NmapPortStatus.DOWN
                protocol = TransportProtocol._from_str(child.attrib["protocol"])
                port_id = child.attrib[constants.NMAP_XML.PORT_ID]
                service_name = constants.NMAP_XML.UNKNOWN
                service_version = ""
                service_fp = ""
                for child_2 in list(child.iter()):
                    if child_2.tag == constants.NMAP_XML.STATE:
                        port_status = NmapUtil._parse_nmap_port_status_xml(child_2)
                    elif child_2.tag == constants.NMAP_XML.SERVICE:
                        service_name = NmapUtil._parse_nmap_service_name_xml(child_2)
                        service_version = NmapUtil._parse_nmap_service_version_xml(child_2)
                        service_fp = NmapUtil._parse_nmap_service_fp_xml(child_2)
                    elif child_2.tag == constants.NMAP_XML.SCRIPT:
                        result, brute_vuln = NmapUtil._parse_nmap_script(child_2, port=port_id, protocol=protocol,
                                                                         service=service_name, action=action)
                        if result is not None:
                            if isinstance(result, list) and len(result) > 0 and isinstance(result[0], NmapVuln):
                                vulnerabilities = result
                            elif isinstance(result, list) and len(result) > 0 \
                                    and isinstance(result[0], NmapBruteCredentials):
                                credentials = result
                                if brute_vuln is not None:
                                    vulnerabilities.append(brute_vuln)
                            elif isinstance(result, NmapHttpEnum):
                                http_enum = result
                            elif isinstance(result, NmapHttpGrep):
                                http_grep = result
                            elif isinstance(result, NmapVulscan):
                                vulscan = result
                if port_status == NmapPortStatus.UP:
                    port = NmapPort(port_id=port_id, protocol=protocol, status=port_status, service_name=service_name,
                                    http_enum=http_enum, http_grep=http_grep, vulscan=vulscan,
                                    service_version=service_version, service_fp=service_fp)
                    ports.append(port)
        return ports, vulnerabilities, credentials

    @staticmethod
    def _parse_nmap_service_name_xml(xml_data) -> Union[str, Any]:
        """
        Parses a XML service name element

        :param xml_data: the XML service element
        :return: the name of the service
        """
        return xml_data.attrib[constants.NMAP_XML.NAME]

    @staticmethod
    def _parse_nmap_service_version_xml(xml_data) -> str:
        """
        Parses a XML service element

        :param xml_data: the XML service element
        :return: the version of the service
        """
        version = ""
        if constants.NMAP_XML.VERSION in xml_data.keys():
            version = xml_data.attrib[constants.NMAP_XML.VERSION]
        return version

    @staticmethod
    def _parse_nmap_service_fp_xml(xml_data) -> str:
        """
        Parses a XML service element

        :param xml_data: the XML service element
        :return: the fingerprint of the service
        """
        servicefp = ""
        if constants.NMAP_XML.SERVICEFP in xml_data.keys():
            servicefp = xml_data.attrib[constants.NMAP_XML.SERVICEFP]
        return servicefp

    @staticmethod
    def _parse_nmap_port_status_xml(xml_data) -> NmapPortStatus:
        """
        Parses a XML port status element

        :param xml_data: the XML port status element
        :return: the parsed port status
        """
        port_status = NmapPortStatus.DOWN
        if xml_data.attrib[constants.NMAP_XML.STATE] == constants.NMAP_XML.OPEN_STATE:
            port_status = NmapPortStatus.UP
        return port_status

    @staticmethod
    def _parse_nmap_os_xml(xml_data) -> List[NmapOs]:
        """
        Parses NMAP OS XML element

        :param xml_data: the XML OS element
        :return: Parsed NmapOS
        """
        os_matches = []
        for child in list(xml_data.iter()):
            if child.tag == constants.NMAP_XML.OS_MATCH:
                name = child.attrib[constants.NMAP_XML.NAME]
                accuracy = int(child.attrib[constants.NMAP_XML.ACCURACY])
                t_acc_cmp = 0
                vendor = ""
                osfamily = ""
                for c2 in list(child.iter()):
                    if c2.tag == constants.NMAP_XML.OS_CLASS:
                        t_acc = int(c2.attrib[constants.NMAP_XML.ACCURACY])
                        if t_acc > t_acc_cmp:
                            vendor = c2.attrib[constants.NMAP_XML.VENDOR]
                            osfamily = c2.attrib[constants.NMAP_XML.OS_FAMILY]
                            t_acc_cmp = t_acc

                os_match = NmapOs(name=name, vendor=vendor, osfamily=osfamily, accuracy=accuracy)
                os_matches.append(os_match)
        return os_matches

    @staticmethod
    def _parse_nmap_table_vuln(xml_data, port: int, protocol: TransportProtocol, service: str) -> NmapVuln:
        """
        Parses a Table XML element with vulnerabilities

        :param xml_data: the XML table element
        :param port: the port element parent
        :param protocol: the protocol of the parent
        :param service: the service running on the port
        :return: parsed Nmap vulnerability
        """
        cvss = constants.VULNERABILITIES.default_cvss
        id = ""
        for child in list(xml_data.iter()):
            if child.tag == constants.NMAP_XML.ELEM:
                if constants.NMAP_XML.KEY in child.keys():
                    if child.attrib[constants.NMAP_XML.KEY] == constants.NMAP_XML.CVSS:
                        cvss = float(child.text)
                    elif child.attrib[constants.NMAP_XML.KEY] == constants.NMAP_XML.ID:
                        id = child.text
        vuln = NmapVuln(name=id, port=port, protocol=protocol, cvss=cvss, service=service)
        return vuln

    @staticmethod
    def _parse_nmap_script(xml_data, port: int, protocol: TransportProtocol, service: str,
                           action: EmulationAttackerAction) \
            -> Tuple[Union[List[NmapVuln], List[NmapBruteCredentials], NmapHttpEnum, NmapHttpGrep, NmapVulscan, None],
                     Union[NmapVuln, None]]:
        """
        Parses a XML script element

        :param xml_data: the XML script element
        :param port: the port of the parent element
        :param protocol: the protocol of the parent element
        :param service: the service running on the port
        :param action: action of the scan
        :return: a list of parsed nmap vulnerabilities or a list of parsed credentials and maybe a vuln
        """
        if constants.NMAP_XML.ID in xml_data.keys():
            if xml_data.attrib[constants.NMAP_XML.ID] == constants.NMAP_XML.VULNERS_SCRIPT_ID:
                return NmapUtil._parse_nmap_vulners(xml_data, port=port, protocol=protocol, service=service), None
            elif xml_data.attrib[constants.NMAP_XML.ID] in constants.NMAP_XML.BRUTE_SCRIPTS:
                return NmapUtil._parse_nmap_brute(xml_data, port=port, protocol=protocol, service=service,
                                                  action=action)
            elif xml_data.attrib[constants.NMAP_XML.ID] == constants.NMAP_XML.HTTP_ENUM_SCRIPT:
                return NmapUtil._parse_nmap_http_enum_xml(xml_data), None
            elif xml_data.attrib[constants.NMAP_XML.ID] == constants.NMAP_XML.HTTP_GREP_SCRIPT:
                return NmapUtil._parse_nmap_http_grep_xml(xml_data), None
            elif xml_data.attrib[constants.NMAP_XML.ID] == constants.NMAP_XML.VULSCAN_SCRIPT:
                return NmapUtil._parse_nmap_http_vulscan_xml(xml_data), None
        return None, None

    @staticmethod
    def _parse_nmap_vulners(xml_data, port: int, protocol: TransportProtocol, service: str) -> List[NmapVuln]:
        """
        Parses a XML result from a vulners scan

        :param xml_data: the XML script element
        :param port: the port of the parent element
        :param protocol: the protocol of the parent element
        :param service: the service running on the port
        :return: a list of parsed nmap vulnerabilities
        """
        vulnerabilities = []
        for child in list(xml_data.iter())[1:]:
            if child.tag == constants.NMAP_XML.TABLE:
                for c_2 in list(child.iter())[1:]:
                    if c_2.tag == constants.NMAP_XML.TABLE:
                        vuln = NmapUtil._parse_nmap_table_vuln(c_2, port=port, protocol=protocol, service=service)
                        vulnerabilities.append(vuln)
                break
        return vulnerabilities

    @staticmethod
    def _parse_nmap_brute(xml_data, port: int, protocol: TransportProtocol, service: str,
                          action: EmulationAttackerAction) -> Tuple[List[NmapBruteCredentials], NmapVuln]:
        """
        Parses a XML result from a brute force dictionary scan

        :param xml_data: the XML script element
        :param port: the port of the parent element
        :param protocol: the protocol of the parent element
        :param service: the service running on the port
        :return: a list of found credentials, vulnerability
        """
        credentials = []
        for child in list(xml_data.iter())[1:]:
            if child.tag == constants.NMAP_XML.TABLE:
                if constants.NMAP_XML.KEY in child.keys():
                    if child.attrib[constants.NMAP_XML.KEY] == constants.NMAP_XML.ACCOUNTS:
                        for c_2 in list(child.iter())[1:]:
                            if c_2.tag == constants.NMAP_XML.TABLE:
                                cred = NmapUtil._parse_nmap_table_cred(c_2, port=port, protocol=protocol,
                                                                       service=service)
                                credentials.append(cred)
                        break
        vulnerability = None
        if len(credentials) > 0:
            vuln_name = EnvDynamicsUtil.exploit_get_vuln_name(a=action)
            service_name = EnvDynamicsUtil.exploit_get_service_name(a=action)
            credentials_1 = list(map(lambda x: x.to_obs(), credentials))
            vulnerability = NmapVuln(name=vuln_name, port=port, protocol=protocol,
                                     cvss=EnvDynamicsUtil.exploit_get_vuln_cvss(a=action),
                                     service=service_name, credentials=credentials_1)
        return credentials, vulnerability

    @staticmethod
    def _parse_nmap_table_cred(xml_data, port: int, protocol: TransportProtocol, service: str) -> NmapBruteCredentials:
        """
        Parses a Table XML element with credentials

        :param xml_data: the XML table element
        :param port: the port element parent
        :param protocol: the protocol of the parent
        :param service: the service running on the port
        :return: parsed Nmap credentials
        """
        username = ""
        pw = ""
        state = ""
        for child in list(xml_data.iter()):
            if child.tag == constants.NMAP_XML.ELEM:
                if constants.NMAP_XML.KEY in child.keys():
                    if child.attrib[constants.NMAP_XML.KEY] == constants.NMAP_XML.USERNAME:
                        username = child.text
                    elif child.attrib[constants.NMAP_XML.KEY] == constants.NMAP_XML.PASSWORD:
                        pw = child.text
                    elif child.attrib[constants.NMAP_XML.KEY] == constants.NMAP_XML.STATE:
                        state = child.text
        credentials = NmapBruteCredentials(username=username, pw=pw, state=state, port=int(port), protocol=protocol,
                                           service=service)
        return credentials

    @staticmethod
    def merge_nmap_scan_result_with_state(scan_result: NmapScanResult, s: EmulationEnvState,
                                          a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Merges a NMAP scan result with an existing observation state

        :param scan_result: the scan result
        :param s: the current state
        :param a: the action just executed
        :return: s'
        """

        new_m_obs = []
        for host in scan_result.hosts:
            m_obs = EmulationAttackerMachineObservationState.from_nmap_result(host)
            # m_obs = EnvDynamicsUtil.brute_tried_flags(a=a, m_obs=m_obs)
            new_m_obs.append(m_obs)

        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, new_m_obs, action=a, emulation_env_config=s.emulation_env_config)

        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime

    @staticmethod
    def nmap_scan_action_helper(s: EmulationEnvState, a: EmulationAttackerAction,
                                masscan: bool = False) -> EmulationEnvState:
        """
        Helper function for executing a NMAP scan action on the emulation. Implements caching.

        :param s: the current env state
        :param a: the NMAP action to execute
        :param masscan: whether it is a masscan or not
        :return: s'
        """
        # If cache miss, then execute cmd
        cmds, file_names = a.nmap_cmds()
        if masscan:
            cmds = a.masscan_cmds()

        Logger.__call__().get_logger().info(
            f"Running NMAP scan on container: {s.emulation_env_config.containers_config.agent_ip}, commands: "
            f"{','.join(cmds)}")
        results = EmulationUtil.execute_ssh_cmds(cmds=cmds, conn=s.emulation_env_config.get_hacker_connection())
        total_time = sum(list(map(lambda x: x[2], results)))

        EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                               emulation_env_config=s.emulation_env_config)

        # Read results
        scan_result = NmapScanResult(hosts=[], ips=[s.emulation_env_config.containers_config.agent_ip])
        for file_name in file_names:
            for i in range(constants.ENV_CONSTANTS.NUM_RETRIES):
                xml_data = NmapUtil.parse_nmap_scan(file_name=file_name,
                                                    emulation_env_config=s.emulation_env_config)
                scan_result_new = NmapUtil.parse_nmap_scan_xml(
                    xml_data, ips=[s.emulation_env_config.containers_config.agent_ip], action=a)
                s.attacker_obs_state.agent_reachable.update(scan_result.reachable)
                scan_result = NmapUtil.merge_nmap_scan_results(scan_result, scan_result_new)
        return NmapUtil.nmap_pivot_scan_action_helper(s=s, a=a, partial_result=scan_result.copy())

    @staticmethod
    def _merge_nmap_hosts(host: NmapHostResult, hosts: List[NmapHostResult]) -> List[NmapHostResult]:
        """
        Merge nmap host results

        :param host: a host to merge with an existing list of hosts
        :param hosts: the list of hosts to merge with the new host
        :return: the merged list
        """
        found = False
        for h in hosts:
            if h.ips_match(host.ips):
                found = True
                vulnerabilities = list(set(h.vulnerabilities).union(host.vulnerabilities))
                h.vulnerabilities = vulnerabilities
                ports = list(set(h.ports).union(host.ports))
                h.ports = ports
        if not found:
            hosts.append(host)
        return hosts

    @staticmethod
    def _parse_nmap_trace_xml(xml_data) -> NmapTrace:
        """
        Parses a trace XML element in the XML tree

        :param xml_data: the trace XML element
        :return: NmapTrace
        """
        hops = []
        for child in list(xml_data.iter()):
            if child.tag == constants.NMAP_XML.HOP:
                hop = NmapUtil._parse_nmap_hop_xml(child)
                hops.append(hop)
        nmap_trace = NmapTrace(hops=hops)
        return nmap_trace

    @staticmethod
    def _parse_nmap_hop_xml(xml_data) -> NmapHop:
        """
        Parses a hop XML element in the XML tree

        :param xml_data: the hop XML element
        :return: NmapHop
        """
        ttl = 0
        ip = ""
        rtt = 0.0
        host = ""
        if constants.NMAP_XML.IPADDR in xml_data.keys():
            ip = xml_data.attrib[constants.NMAP_XML.IPADDR]
        if constants.NMAP_XML.RTT in xml_data.keys():
            rtt = float(xml_data.attrib[constants.NMAP_XML.RTT])
        if constants.NMAP_XML.TTL in xml_data.keys():
            ttl = int(xml_data.attrib[constants.NMAP_XML.TTL])
        if constants.NMAP_XML.HOST in xml_data.keys():
            host = xml_data.attrib[constants.NMAP_XML.HOST]
        nmap_hop = NmapHop(ttl=ttl, ipaddr=ip, rtt=rtt, host=host)
        return nmap_hop

    @staticmethod
    def _parse_nmap_http_enum_xml(xml_data) -> NmapHttpEnum:
        """
        Parses a http enum XML element in the XML tree

        :param xml_data: the http enum XML element
        :return: HttpEnum
        """
        output = ""
        if constants.NMAP_XML.OUTPUT in xml_data.keys():
            output = xml_data.attrib[constants.NMAP_XML.OUTPUT]
        nmap_http_enum = NmapHttpEnum(output=output)
        return nmap_http_enum

    @staticmethod
    def _parse_nmap_http_grep_xml(xml_data) -> NmapHttpGrep:
        """
        Parses a http grep XML element in the XML tree

        :param xml_data: the http grep XML element
        :return: NmapHttpGrep
        """
        output = ""
        if constants.NMAP_XML.OUTPUT in xml_data.keys():
            output = xml_data.attrib[constants.NMAP_XML.OUTPUT]
        nmap_http_grep = NmapHttpGrep(output=output)
        return nmap_http_grep

    @staticmethod
    def _parse_nmap_http_vulscan_xml(xml_data) -> NmapVulscan:
        """
        Parses a vulscan XML element in the XML tree

        :param xml_data: the vulscan XML element
        :return: NmapVulScan
        """
        output = ""
        if constants.NMAP_XML.OUTPUT in xml_data.keys():
            output = xml_data.attrib[constants.NMAP_XML.OUTPUT]
        nmap_vulscan = NmapVulscan(output=output)
        return nmap_vulscan

    @staticmethod
    def merge_nmap_scan_results(scan_result_1: NmapScanResult, scan_result_2: NmapScanResult) -> NmapScanResult:
        """
        Merges two nmap scan results

        :param scan_result_1: the first result to merge
        :param scan_result_2: the second result to merge
        :return: the merged result
        """
        new_hosts = []

        for h in scan_result_2.hosts:
            new_host = True
            for h2 in scan_result_1.hosts:
                if h.ips_match(h2.ips):
                    new_host = False
            if new_host:
                new_hosts.append(h)

        for h in scan_result_1.hosts:
            for h2 in scan_result_2.hosts:
                if h.ips_match(h2.ips):
                    h.hostnames = list(set(h.hostnames).union(set(h2.hostnames)))
                    h.ports = list(set(h.ports).union(h2.ports))
                    h.vulnerabilities = list(set(h.vulnerabilities).union(h2.vulnerabilities))
                    h.credentials = list(set(h.credentials).union(h2.credentials))
                    if h.os is None:
                        h.os = h2.os
                    if h.trace is None:
                        h.trace = h2.trace

        scan_result_1.hosts = scan_result_1.hosts + new_hosts
        scan_result_1.reachable = list(set(scan_result_1.reachable + scan_result_2.reachable))
        return scan_result_1

    @staticmethod
    def nmap_pivot_scan_action_helper(s: EmulationEnvState, a: EmulationAttackerAction,
                                      partial_result: NmapScanResult) -> EmulationEnvState:
        """
        Performs an NMAP pivot scan, utilizing many compromised hosts

        :param s: the curretn state
        :param a: the attacker scan action
        :param partial_result: the initial result before pivoting
        :return: the new state
        """
        merged_scan_result = partial_result
        total_results: List[NmapScanResult] = []
        threads = []
        for machine in s.attacker_obs_state.machines:
            if machine.logged_in and machine.tools_installed and machine.backdoor_installed:
                thread = PivotNMAPScanThread(machine=machine, s=s, a=a)
                thread.start()
                threads.append(thread)
        for thread in threads:
            thread.join()
        for i in range(len(threads)):
            if merged_scan_result is not None and threads[i].scan_result is not None:
                total_results.append(threads[i].scan_result)
                merged_scan_result = NmapUtil.merge_nmap_scan_results(scan_result_1=merged_scan_result,
                                                                      scan_result_2=threads[i].scan_result.copy())
            elif merged_scan_result is None:
                total_results.append(threads[i].scan_result)
                merged_scan_result = threads[i].scan_result.copy()

        s_prime = NmapUtil.merge_nmap_scan_result_with_state(scan_result=merged_scan_result, s=s, a=a)
        new_machines_obs_1 = []
        reachable = s.attacker_obs_state.agent_reachable
        reachable.add(s.emulation_env_config.containers_config.router_ip)

        for machine in s_prime.attacker_obs_state.machines:
            if machine.logged_in and machine.tools_installed and machine.backdoor_installed:
                reachable = reachable.union(machine.reachable)

        for machine in s_prime.attacker_obs_state.machines:
            if machine.logged_in and machine.tools_installed:
                machine = EnvDynamicsUtil.ssh_backdoor_tried_flags(a=a, m_obs=machine)

            if machine.ips_match(reachable) and (machine.ips_match(a.ips) or a.index == -1):
                machine = EnvDynamicsUtil.exploit_tried_flags(a=a, m_obs=machine)

            valid_ips = True
            for ip in machine.ips:
                if int(ip.split(".")[-1]) == 1:
                    valid_ips = False
            if valid_ips:
                new_machines_obs_1.append(machine)
        s_prime.attacker_obs_state.machines = new_machines_obs_1
        s_prime.attacker_obs_state.sort_machines()
        s_prime.defender_obs_state.sort_machines()
        return s_prime


class PivotNMAPScanThread(threading.Thread):
    """
    Thread for asynchronous nmap scanning
    """

    def __init__(self, machine: EmulationAttackerMachineObservationState, a: EmulationAttackerAction,
                 s: EmulationEnvState):
        """
        Initializes the thread

        :param machine: the machine to do the details on
        :param a: the scanning action
        :param s: the current environment state
        """
        threading.Thread.__init__(self)
        self.machine = machine
        self.a = a
        self.s = s
        self.total_time = 0
        self.scan_result: NmapScanResult = NmapScanResult(hosts=[], ips=[])

    def run(self) -> None:
        """
        Main loop of the thread; performs the scanning

        :return: None
        """
        ssh_connections_alive = []
        for c in self.machine.ssh_connections:
            try:
                EmulationUtil.execute_ssh_cmds(cmds=[constants.COMMANDS.LS], conn=c.conn)
                ssh_connections_alive.append(c)
            except Exception as e:
                Logger.__call__().get_logger().debug(f"There was an error connecting to {str(e)}, {repr(e)}")
                new_conn = ConnectionUtil.reconnect_ssh(c)
                ssh_connections_alive.append(new_conn)
        self.machine.ssh_connections = ssh_connections_alive
        ssh_connections_sorted_by_root = sorted(
            self.machine.ssh_connections,
            key=lambda x: (constants.SSH_BACKDOOR.BACKDOOR_PREFIX in x.credential.username, x.root,
                           x.credential.username),
            reverse=True)
        for c in ssh_connections_sorted_by_root:
            cwd = "/home/" + c.credential.username + "/"
            cmds, file_names = self.a.nmap_cmds(machine_ips=self.machine.ips)
            results = []
            c2 = c
            for i in range(constants.ENV_CONSTANTS.NUM_RETRIES):
                try:
                    Logger.__call__().get_logger().info(
                        f"Running NMAP scan on container(s): {self.machine.ips}, commands: "
                        f"{','.join(cmds)}")
                    results = EmulationUtil.execute_ssh_cmds(cmds=cmds, conn=c2.conn)
                    break
                except Exception as e:
                    Logger.__call__().get_logger().warning(
                        f"exception execution commands for ip:{c.ip}, "
                        f"username: {c.credential.username}, conn: {c.conn}, "
                        f"transport: {c.conn.get_transport()}, active: {c.conn.get_transport().is_active()},"
                        f"{str(e)}, {repr(e)}")
                    c2 = ConnectionUtil.reconnect_ssh(c2)
                    time.sleep(constants.ENV_CONSTANTS.SLEEP_RETRY)
            total_time = sum(list(map(lambda x: x[2], results)))
            EmulationUtil.log_measured_action_time(
                total_time=total_time, action=self.a, emulation_env_config=self.s.emulation_env_config)

            # Read result
            scan_result = NmapScanResult(hosts=[], ips=self.machine.ips)
            for file_name in file_names:
                for i in range(constants.ENV_CONSTANTS.NUM_RETRIES):
                    try:
                        xml_data = NmapUtil.parse_nmap_scan(
                            file_name=file_name, emulation_env_config=self.s.emulation_env_config,
                            conn=c.conn, dir=cwd)
                        new_scan_result = NmapUtil.parse_nmap_scan_xml(xml_data, ips=self.machine.ips, action=self.a)
                        scan_result = NmapUtil.merge_nmap_scan_results(scan_result_1=scan_result,
                                                                       scan_result_2=new_scan_result)
                        self.machine.reachable.update(scan_result.reachable)
                    except Exception as e:
                        Logger.__call__().get_logger().warning(
                            f"There was an exception parsing the file:{file_name} on ip:{c.ip}, error:{e}")
                        time.sleep(constants.ENV_CONSTANTS.SLEEP_RETRY)
            self.scan_result = scan_result
            break



from typing import Tuple
import time
import random
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.credential import Credential
from csle_common.util.emulation_util import EmulationUtil
from csle_common.util.connection_util import ConnectionUtil
from csle_common.logging.log import Logger


class ShellUtil:
    """
    Class containing utility functions for the shell-related functionality to the emulation
    """

    @staticmethod
    def _find_flag_using_ssh(machine: EmulationAttackerMachineObservationState,
                             emulation_env_config: EmulationEnvConfig,
                             a: EmulationAttackerAction, new_m_obs: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, float, bool]:
        """
        Utility function for using existing SSH connections to a specific machine to search the file system for flags

        :param machine: the machine to search
        :param emulation_env_config: the emulation env config
        :param a: the action of finding the flags
        :param new_m_obs: the updated machine observation with the found flags
        :return: the updated machine observation with the found flags, cost, root
        """
        total_cost = 0
        ssh_connections_sorted_by_root = sorted(
            machine.ssh_connections,
            key=lambda x: (constants.SSH_BACKDOOR.BACKDOOR_PREFIX in x.credential.username, x.root,
                           x.credential.username),
            reverse=True)
        root_scan = False
        flag_paths = []
        for c in ssh_connections_sorted_by_root:
            cmd = a.cmds[0]
            if c.root:
                cmd = constants.COMMANDS.SUDO + " " + cmd
            for i in range(constants.ENV_CONSTANTS.ATTACKER_SSH_RETRY_FIND_FLAG):
                outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=c.conn)
                new_m_obs.filesystem_searched = True
                EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                                       emulation_env_config=emulation_env_config)
                outdata_str = outdata.decode()
                flag_paths = outdata_str.split("\n")
                flag_paths = list(filter(lambda x: x != '', flag_paths))
                num_flags = 0
                for fp in flag_paths:
                    fp = fp.replace(".txt", "")
                    for node_flags_config in emulation_env_config.flags_config.node_flag_configs:
                        if node_flags_config.ip in machine.ips:
                            for flag in node_flags_config.flags:
                                if flag.name == fp:
                                    num_flags += 1
                    if fp in emulation_env_config.flags_config.node_flag_configs:
                        num_flags += 1
                if len(flag_paths) > 0 and num_flags > 0:
                    break
                else:
                    time.sleep(1)

            # Check for flags
            for fp in flag_paths:
                fp = fp.replace(".txt", "")
                for node_flags_config in emulation_env_config.flags_config.node_flag_configs:
                    if node_flags_config.ip in machine.ips:
                        for flag in node_flags_config.flags:
                            if flag.name == fp:
                                new_m_obs.flags_found.add(flag)

            if c.root:
                root_scan = True
                break
        return new_m_obs, total_cost, root_scan

    @staticmethod
    def _find_flag_using_telnet(machine: EmulationAttackerMachineObservationState,
                                emulation_env_config: EmulationEnvConfig, a: EmulationAttackerAction,
                                new_m_obs: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, float, bool]:
        """
        Utility function for using existing Telnet connections to a specific machine to search the file system for flags

        :param machine: the machine to search
        :param emulation_env_config: the emulation env config
        :param a: the action of finding the flags
        :param new_m_obs: the updated machine observation with the found flags
        :return: the updated machine observation with the found flags, cost, root
        """
        total_cost = 0
        telnet_connections_sorted_by_root = sorted(
            machine.telnet_connections,
            key=lambda x: (constants.SSH_BACKDOOR.BACKDOOR_PREFIX in x.credential.username, x.root,
                           x.credential.username),
            reverse=True)
        root_scan = False
        for c in telnet_connections_sorted_by_root:
            cmd = a.cmds[0] + "\n"
            if c.root:
                cmd = constants.COMMANDS.SUDO + " " + cmd
            start = time.time()
            c.conn.write(cmd.encode())
            response = c.conn.read_until(constants.TELNET.PROMPT, timeout=5)
            new_m_obs.filesystem_searched = True
            end = time.time()
            total_time = end - start

            EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                                   emulation_env_config=emulation_env_config)
            flag_paths = response.decode().strip().split("\r\n")
            # Check for flags
            for fp in flag_paths:
                fp = fp.replace(".txt", "")
                for node_flags_config in emulation_env_config.flags_config.node_flag_configs:
                    if node_flags_config.ip in machine.ips:
                        for flag in node_flags_config.flags:
                            if flag.name == fp:
                                new_m_obs.flags_found.add(flag)

            if c.root:
                root_scan = True
                break
        return new_m_obs, total_cost, root_scan

    @staticmethod
    def _find_flag_using_ftp(machine: EmulationAttackerMachineObservationState,
                             emulation_env_config: EmulationEnvConfig, a: EmulationAttackerAction,
                             new_m_obs: EmulationAttackerMachineObservationState) \
            -> Tuple[EmulationAttackerMachineObservationState, float, bool]:
        """
        Utility function for using existing FTP connections to a specific machine to search the file system for flags

        :param machine: the machine to search
        :param emulation_env_config: the emulation env config
        :param a: the action of finding the flags
        :param new_m_obs: the updated machine observation with the found flags
        :return: the updated machine observation with the found flags, cost, root
        """
        total_cost = 0
        ftp_connections_sorted_by_root = sorted(
            machine.ftp_connections,
            key=lambda x: (constants.SSH_BACKDOOR.BACKDOOR_PREFIX in x.credential.username, x.root,
                           x.credential.username),
            reverse=True)
        root_scan = False
        flag_paths = []
        for c in ftp_connections_sorted_by_root:
            for i in range(constants.ENV_CONSTANTS.ATTACKER_FTP_RETRY_FIND_FLAG):
                cmd = a.alt_cmds[0] + "\n"
                if c.root:
                    cmd = constants.COMMANDS.SUDO + " " + cmd
                start = time.time()
                c.interactive_shell.send(cmd)
                output = b""

                # clear output
                if c.interactive_shell.recv_ready():
                    c.interactive_shell.recv(constants.COMMON.DEFAULT_RECV_SIZE)
                command_complete = False
                timeouts = 0
                while not command_complete:
                    while not c.interactive_shell.recv_ready():
                        if timeouts > constants.ENV_CONSTANTS.SHELL_MAX_TIMEOUTS:
                            Logger.__call__().get_logger().warning("max timeouts FTP, env:{}".format(
                                emulation_env_config.containers_config.agent_ip))
                            break
                        time.sleep(constants.ENV_CONSTANTS.SHELL_READ_WAIT)
                        timeouts += 1
                    if c.interactive_shell.recv_ready():
                        output += c.interactive_shell.recv(constants.COMMON.LARGE_RECV_SIZE)
                        timeouts = 0
                        if constants.FTP.LFTP_PROMPT in output.decode() \
                                or constants.FTP.LFTP_PROMPT_2 in output.decode():
                            command_complete = True
                            end = time.time()
                            total_time = end - start

                            EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                                                   emulation_env_config=emulation_env_config)
                    else:
                        break

                    output_str = output.decode("utf-8")
                    output_str = constants.NMAP.SHELL_ESCAPE.sub("", output_str)
                    output_list = output_str.split('\r\n')
                    output_list = output_list[1:-1]  # remove command ([0]) and prompt ([-1])
                    flag_paths = list(filter(lambda x: constants.FTP.ACCESS_FAILED not in x and x != "", output_list))
                    ff = False
                    # Check for flags
                    for fp in flag_paths:
                        fp = fp.replace(".txt", "")
                        for node_flags_config in emulation_env_config.flags_config.node_flag_configs:
                            if node_flags_config.ip in machine.ips:
                                for flag in node_flags_config.flags:
                                    if flag.name == fp:
                                        ff = True
                    if not ff:
                        continue
                    else:
                        break
            new_m_obs.filesystem_searched = True

            # Check for flags
            for fp in flag_paths:
                fp = fp.replace(".txt", "")
                for node_flags_config in emulation_env_config.flags_config.node_flag_configs:
                    if node_flags_config.ip in machine.ips:
                        for flag in node_flags_config.flags:
                            if flag.name == fp:
                                new_m_obs.flags_found.add(flag)

            if c.root:
                root_scan = True
                break
        return new_m_obs, total_cost, root_scan

    @staticmethod
    def parse_tools_installed_file(file_name: str, emulation_env_config: EmulationEnvConfig) -> bool:
        """
        Parses a file containing cached results of a install-tools action

        :param file_name: name of the file to parse
        :param emulation_env_config: environment config
        :return: boolean: if installed or not
        """
        sftp_client = emulation_env_config.get_hacker_connection().open_sftp()
        remote_file = sftp_client.open(constants.NMAP.RESULTS_DIR + file_name)
        try:
            data = remote_file.read()
            data = data.decode()
            installed = bool(int(data))
        finally:
            remote_file.close()
        return installed

    @staticmethod
    def _parse_tools_installed_check_result(result: str) -> bool:
        """
        Checks the output result of a tools install action to see whether the action was successful or not.

        :param result: the result to check
        :return: True if sucessful otherwise False
        """
        return ("will be installed" in result or "already installed" in result
                or "already the newest version" in result)

    @staticmethod
    def install_tools_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Uses compromised machines with root access to install tools

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        new_machines_obs = []
        total_cost = 0.0
        total_time = 0.0
        for machine in s.attacker_obs_state.machines:
            new_m_obs = EmulationAttackerMachineObservationState(ips=machine.ips)
            if machine.logged_in and machine.root:
                new_m_obs.install_tools_tried = True
            installed = False
            if machine.logged_in and machine.root and not machine.tools_installed:
                # Start with ssh connections
                ssh_root_connections = list(filter(lambda x: x.root, machine.ssh_connections))
                ssh_root_connections = list(sorted(ssh_root_connections, key=lambda x: x.credential.username))
                ssh_cost = 0.0
                for c in ssh_root_connections:
                    cmd = a.cmds[0]
                    for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_INSTALL_TOOLS):
                        outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=c.conn)
                        time.sleep(constants.ENV_CONSTANTS.ATTACKER_INSTALL_TOOLS_SLEEP_SECONDS)
                        outdata = outdata.decode()
                        ssh_cost += float(total_time)
                        if ShellUtil._parse_tools_installed_check_result(result=outdata):
                            installed = True
                            new_m_obs.tools_installed = True
                        else:
                            Logger.__call__().get_logger().warning(
                                "SSH tools installed failed result. out:{}, err:{}".format(outdata, errdata))
                        if installed:
                            break

                    # try to download seclists
                    seclists_installed = ShellUtil._check_if_seclists_is_installed(conn=c.conn, telnet=False)
                    if not seclists_installed:
                        cmd = a.cmds[1]
                        outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=c.conn)
                        ssh_cost += float(total_time)

                    EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                                           emulation_env_config=s.emulation_env_config)

                    new_machines_obs.append(new_m_obs)

                    if installed:
                        break

                total_cost += ssh_cost

                # Telnet connections
                telnet_cost = 0.0
                if installed:
                    continue
                telnet_root_connections = list(filter(lambda x: x.root, machine.telnet_connections))
                telnet_root_connections = list(sorted(telnet_root_connections, key=lambda x: x.credential.username))
                for c in telnet_root_connections:
                    # Install packages
                    cmd = a.cmds[0] + "\n"
                    start = time.time()
                    for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_INSTALL_TOOLS):
                        try:
                            c.conn.write(cmd.encode())
                            response = c.conn.read_until(constants.TELNET.PROMPT, timeout=25)
                            response = response.decode()
                            end = time.time()
                            total_time = end - start
                            telnet_cost += float(total_time)
                            if ShellUtil._parse_tools_installed_check_result(result=response):
                                installed = True
                                new_m_obs.tools_installed = True
                            else:
                                Logger.__call__().get_logger().warning(
                                    "Telnet tools installed failed result.{}".format(response))
                            if installed:
                                break
                        except Exception as e:
                            Logger.__call__().get_logger().warning(
                                f"Telnet tools installed exception {str(e)}, {repr(e)}")
                            c = ConnectionUtil.reconnect_telnet(
                                c=c, forward_port=s.emulation_env_config.get_port_forward_port())

                    seclists_installed = ShellUtil._check_if_seclists_is_installed(conn=c.conn, telnet=True)
                    if not seclists_installed:
                        # Try to download SecLists
                        cmd = a.cmds[1] + "\n"
                        start = time.time()
                        c.conn.write(cmd.encode())
                        response = c.conn.read_until(constants.TELNET.PROMPT, timeout=2000)
                        response = response.decode()
                        end = time.time()
                        total_time = end - start
                        telnet_cost += float(total_time)

                    EmulationUtil.log_measured_action_time(total_time=total_time, action=a,
                                                           emulation_env_config=s.emulation_env_config)

                    new_machines_obs.append(new_m_obs)

                    if installed:
                        break

                total_cost += telnet_cost
        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, new_machines_obs, emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations

        return s_prime

    @staticmethod
    def _check_if_seclists_is_installed(conn, telnet: bool = False) -> bool:
        """
        Checks if seclists are downloaded

        :param conn: the connection to use for the command
        :param telnet: whether the connection is a telnet connection
        :return: True if downloaded, else false
        """
        cmd = constants.SHELL.CHECK_FOR_SECLISTS
        checklists_installed = False
        if not telnet:
            for i in range(8):
                outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=conn)
                checklists_installed = "file exists" in outdata.decode() or "file exists" in errdata.decode()
                if checklists_installed:
                    break
                else:
                    Logger.__call__().get_logger().warning(
                        "checklists not installed:{}, {}".format(outdata.decode(), errdata.decode()))
            return checklists_installed
        else:
            cmd = cmd + "\n"
            conn.write(cmd.encode())
            response = conn.read_until(constants.TELNET.PROMPT, timeout=5)
            return "file exists" in response.decode()

    @staticmethod
    def execute_ssh_backdoor_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Uses compromised machines with root access to setup SSH backdoor

        :param s: the current state
        :param a: the action to take
        :return: s_prime
        """
        username = constants.SSH_BACKDOOR.BACKDOOR_PREFIX + "_" + str(random.randint(0, 100000))
        pw = constants.SSH_BACKDOOR.DEFAULT_PW
        new_machines_obs = []
        total_cost = 0.0
        for machine in s.attacker_obs_state.machines:
            new_m_obs = EmulationAttackerMachineObservationState(ips=machine.ips)
            backdoor_created = False
            if machine.logged_in and machine.root and machine.tools_installed and not machine.backdoor_installed:
                new_m_obs.backdoor_tried = True
                # Check cached connections
                for cr in s.attacker_cached_backdoor_credentials.values():
                    for ip in machine.ips:
                        if (ip, cr.username, cr.kafka_port) in s.attacker_cached_ssh_connections:
                            conn_dto = s.attacker_cached_ssh_connections[
                                (ip, cr.username, cr.kafka_port)]
                            connection_dto = EmulationConnectionObservationState(
                                conn=conn_dto.conn, credential=cr, root=machine.root,
                                service=constants.SSH.SERVICE_NAME, port=cr.kafka_port, ip=ip)
                            new_m_obs.shell_access_credentials.append(cr)
                            new_m_obs.backdoor_credentials.append(cr)
                            new_m_obs.ssh_connections.append(connection_dto)
                            new_m_obs.backdoor_installed = True
                            new_machines_obs.append(new_m_obs)
                            backdoor_created = True
                            break

                if backdoor_created:
                    continue

                # Try first to setup new ssh connections
                ssh_root_connections = list(filter(lambda x: x.root, machine.ssh_connections))
                ssh_root_connections = sorted(ssh_root_connections, key=lambda x: x.credential.username)
                ssh_cost = 0.0
                for c in ssh_root_connections:
                    users = EmulationUtil._list_all_users(c, emulation_env_config=s.emulation_env_config)
                    users = sorted(users, key=lambda x: x)
                    user_exists = False
                    for user in users:
                        if constants.SSH_BACKDOOR.BACKDOOR_PREFIX in user and not user == constants.SAMBA.BACKDOOR_USER:
                            user_exists = True
                            username = user

                    if not user_exists:
                        # Create user
                        create_user_cmd = a.cmds[1].format(username, pw, username)
                        outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=create_user_cmd, conn=c.conn)
                        ssh_cost += float(total_time)

                    credential = Credential(username=username, pw=pw, port=22, service="ssh")

                    # Start SSH Server
                    ssh_running = EmulationUtil._check_if_ssh_server_is_running(c.conn)
                    if not ssh_running:
                        start_ssh_cmd = a.cmds[0]
                        outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=start_ssh_cmd, conn=c.conn)
                        ssh_cost += float(total_time)

                    # Create SSH connection
                    new_m_obs.shell_access_credentials.append(credential)
                    new_m_obs.backdoor_credentials.append(credential)
                    a.ips = machine.ips
                    setup_connection_dto = None
                    for i in range(5):
                        setup_connection_dto = ConnectionUtil._ssh_setup_connection(
                            a=a, credentials=[credential], proxy_connections=[c.proxy], s=s)
                        ssh_cost += setup_connection_dto.total_time
                        if len(setup_connection_dto.target_connections) > 0:
                            break
                        else:
                            time.sleep(5)
                    if setup_connection_dto is None:
                        Logger.__call__().get_logger().warning(f"Cannot connect to {a.ips}")
                    else:
                        if len(setup_connection_dto.target_connections) == 0:
                            Logger.__call__().get_logger().warning(f"cannot install backdoor, machine:{machine.ips}, "
                                                                   f"credentials:{credential}")
                        connection_dto = EmulationConnectionObservationState(
                            conn=setup_connection_dto.target_connections[0], credential=credential, root=machine.root,
                            service=constants.SSH.SERVICE_NAME, port=credential.port,
                            proxy=setup_connection_dto.proxies[0],
                            ip=setup_connection_dto.ip)
                        new_m_obs.ssh_connections.append(connection_dto)
                        new_m_obs.backdoor_installed = True
                        new_machines_obs.append(new_m_obs)
                        backdoor_created = True

                        if backdoor_created:
                            break

                total_cost += ssh_cost

                # Telnet connections
                telnet_cost = 0.0
                if backdoor_created:
                    continue
                telnet_root_connections = list(filter(lambda x: x.root, machine.telnet_connections))
                telnet_root_connections = list(sorted(telnet_root_connections, key=lambda x: x.credential.username))
                for c in telnet_root_connections:
                    try:
                        users = EmulationUtil._list_all_users(c,
                                                              emulation_env_config=s.emulation_env_config, telnet=True)
                        user_exists = False
                        for user in users:
                            if constants.SSH_BACKDOOR.BACKDOOR_PREFIX in user \
                                    and not user == constants.SAMBA.BACKDOOR_USER:
                                user_exists = True
                                username = user

                        credential = Credential(username=username, pw=pw, port=22, service="ssh")

                        if not user_exists:
                            # Create user
                            create_user_cmd = a.cmds[1].format(username, pw, username) + "\n"
                            c.conn.write(create_user_cmd.encode())
                            c.conn.read_until(constants.TELNET.PROMPT, timeout=5)

                        ssh_running = EmulationUtil._check_if_ssh_server_is_running(c.conn, telnet=True)
                        if not ssh_running:
                            # Start SSH Server
                            start_ssh_cmd = a.cmds[0] + "\n"
                            c.conn.write(start_ssh_cmd.encode())
                            c.conn.read_until(constants.TELNET.PROMPT, timeout=5)

                        # Create SSH connection
                        new_m_obs.shell_access_credentials.append(credential)
                        new_m_obs.backdoor_credentials.append(credential)
                        a.ips = machine.ips
                        setup_connection_dto = ConnectionUtil._ssh_setup_connection(
                            a=a, credentials=[credential], proxy_connections=[c.proxy], s=s)
                        telnet_cost += setup_connection_dto.total_time
                        connection_dto = EmulationConnectionObservationState(
                            conn=setup_connection_dto.target_connections[0], credential=credential, root=machine.root,
                            service=constants.SSH.SERVICE_NAME, port=credential.port,
                            proxy=setup_connection_dto.proxies[0], ip=setup_connection_dto.ip)
                        new_m_obs.ssh_connections.append(connection_dto)
                        new_m_obs.backdoor_installed = True
                        new_machines_obs.append(new_m_obs)
                        backdoor_created = True
                    except Exception as e:
                        Logger.__call__().get_logger().warning(
                            f"Exception occurred while setting up backdoors , {str(e)}, {repr(e)}")
                    if backdoor_created:
                        break

                total_cost += telnet_cost
        attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
            s.attacker_obs_state.machines, new_machines_obs, emulation_env_config=s.emulation_env_config, action=a)
        s_prime = s
        s_prime.attacker_obs_state.machines = attacker_machine_observations

        return s_prime

    @staticmethod
    def execute_service_login_helper(s: EmulationEnvState, a: EmulationAttackerAction) -> EmulationEnvState:
        """
        Executes a service login on the emulation using previously found credentials

        :param s: the current state
        :param a: the action to take
        :return: s_prime, reward, done
        """
        s_prime = s

        for machine in s.attacker_obs_state.machines:
            a.ips = machine.ips
            s_1, new_conn_ssh = ConnectionUtil.login_service_helper(
                s=s_prime, a=a, alive_check=EmulationEnvConfig.check_if_ssh_connection_is_alive,
                service_name=constants.SSH.SERVICE_NAME)
            s_2, new_conn_ftp = ConnectionUtil.login_service_helper(
                s=s_1, a=a, alive_check=EnvDynamicsUtil.check_if_ftp_connection_is_alive,
                service_name=constants.FTP.SERVICE_NAME)
            s_3, new_conn_telnet = ConnectionUtil.login_service_helper(
                s=s_2, a=a, alive_check=EnvDynamicsUtil.check_if_telnet_connection_is_alive,
                service_name=constants.TELNET.SERVICE_NAME)

            s_prime = s_3

            for m in s_prime.attacker_obs_state.machines:
                if m.ips == a.ips:
                    m.untried_credentials = False

        a.ips = ""

        return s_prime
